# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Here are general and special bokeh plots to use

"""
from masci_tools.vis.bokeh_plotter import BokehPlotter
from masci_tools.vis import ensure_plotter_consistency, NestedPlotParameters

import pandas as pd
from pprint import pprint

################## Helpers     ################

plot_params = BokehPlotter()


def set_bokeh_plot_defaults(**kwargs):
    """
    Set defaults for bokeh backend
    according to the given keyword arguments

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.set_defaults(**kwargs)


def reset_bokeh_plot_defaults():
    """
    Reset the defaults for bokeh backend
    to the hardcoded defaults

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.reset_defaults()


def show_bokeh_plot_defaults():
    """
    Show the currently set defaults for bokeh backend

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    pprint(plot_params.get_dict())


def get_bokeh_help(key):
    """
    Print the decription of the given key in the bokeh backend

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.get_description(key)


##################################### general plots ##########################


@ensure_plotter_consistency(plot_params)
def bokeh_scatter(source,
                  *,
                  xdata='x',
                  ydata='y',
                  xlabel='x',
                  ylabel='y',
                  title='',
                  figure=None,
                  outfilename='scatter.html',
                  **kwargs):
    """
    Create an interactive scatter plot with bokeh

    :param source: source for the data of the plot (pandas Dataframe for example)
    :param xdata: key from which to pull the data for the x-axis
    :param ydata: key from which to pull the data for the y-axis
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `scatter`
    """

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)

    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    res = p.scatter(x=xdata, y=ydata, source=source, **kwargs)

    if plot_params['level'] is not None:
        res.level = plot_params['level']

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.save_plot(p)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_multi_scatter(source,
                        *,
                        xdata='x',
                        ydata='y',
                        figure=None,
                        xlabel='x',
                        ylabel='y',
                        title='',
                        outfilename='scatter.html',
                        **kwargs):
    """
    Create an interactive scatter (muliple data sets possible) plot with bokeh

    :param source: source for the data of the plot (pandas Dataframe for example)
    :param xdata: key from which to pull the data for the x-axis (or if source is None list with data for x-axis)
    :param ydata: key from which to pull the data for the y-axis (or if source is None list with data for y-axis)
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `scatter`
    """
    from bokeh.models import ColumnDataSource

    plot_params.set_defaults(default_type='function', name='scatter plot')

    default_legend_label = ydata
    if source is not None:
        if not isinstance(ydata, list):
            if not isinstance(xdata, list):
                xdata = [xdata]
            ydata = [ydata] * len(xdata)
            default_legend_label = xdata

    if isinstance(xdata, list):
        if len(xdata) != len(ydata):
            xdata = xdata[0]

    plot_params.single_plot = False
    plot_params.num_plots = len(ydata)

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    #Process the given color arguments
    plot_params.set_color_palette_by_num_plots()

    # prepare ColumnDataSource for plot
    if source is None:  # create columndatasources from data given
        # Columns need to have same length
        source = []
        if isinstance(ydata[0], list):
            ydatad = []
            xdatad = []
            for i, ydat in enumerate(ydata):
                label = 'y{}'.format(i)
                ydatad.append(label)
                xdatad.append('x{}'.format(i))
                if isinstance(xdata[0], list):
                    xdat = xdata[i]
                else:
                    xdat = xdata[0]
                source.append(ColumnDataSource({'x': xdat, 'y': ydata}))
        else:
            raise ValueError('If no source dataframe or ColumnData is given, ydata has to be a list'
                             ' of lists, not of type: {}'.format(type(ydata[0])))
    else:
        xdatad = xdata
        ydatad = ydata

    # draw line plot
    # dataframe and column data source expect all entries to be same length...
    # therefore we parse data to plot routines directly... might make other things harder

    plot_kwargs = plot_params.plot_kwargs(plot_type='scatter')

    for indx, data in enumerate(zip(ydatad, plot_kwargs)):

        yname, plot_kw = data

        if isinstance(xdatad, list):
            xdat = xdatad[indx]
        else:
            xdat = xdatad

        if isinstance(source, list):
            sourcet = source[indx]
        else:
            sourcet = source

        if isinstance(default_legend_label, list):
            leg_label = default_legend_label[indx]
        else:
            leg_label = default_legend_label

        if 'legend_label' not in plot_kw:
            plot_kw['legend_label'] = leg_label

        res = p.scatter(x=xdat, y=yname, source=sourcet, **plot_kw)

        if plot_params[('level', indx)] is not None:
            res.level = plot_params[('level', indx)]

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.set_legend(p)
    plot_params.save_plot(p)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_line(source,
               *,
               xdata='x',
               ydata='y',
               figure=None,
               xlabel='x',
               ylabel='y',
               title='',
               outfilename='scatter.html',
               plot_points=False,
               **kwargs):
    """
    Create an interactive multi-line plot with bokeh

    :param source: source for the data of the plot (pandas Dataframe for example)
    :param xdata: key from which to pull the data for the x-axis (or if source is None list with data for x-axis)
    :param ydata: key from which to pull the data for the y-axis (or if source is None list with data for y-axis)
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file
    :param plot_points: bool, if True also plot the points with a scatterplot on top

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `line`
    """
    from bokeh.models import ColumnDataSource

    plot_params.set_defaults(default_type='function', name='line plot')

    default_legend_label = ydata
    if source is not None:
        if not isinstance(ydata, list):
            if not isinstance(xdata, list):
                xdata = [xdata]
            ydata = [ydata] * len(xdata)
            default_legend_label = xdata

    if isinstance(xdata, list):
        if len(xdata) != len(ydata):
            xdata = xdata[0]

    plot_params.single_plot = False
    plot_params.num_plots = len(ydata)

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    #Process the given color arguments
    plot_params.set_color_palette_by_num_plots()

    # prepare ColumnDataSource for plot
    if source is None:  # create columndatasources from data given
        # Columns need to have same length
        source = []
        if isinstance(ydata[0], list):
            ydatad = []
            xdatad = []
            for i, ydat in enumerate(ydata):
                label = 'y{}'.format(i)
                ydatad.append(label)
                xdatad.append('x{}'.format(i))
                if isinstance(xdata[0], list):
                    xdat = xdata[i]
                else:
                    xdat = xdata[0]
                source.append(ColumnDataSource({'x': xdat, 'y': ydata}))
        else:
            raise ValueError('If no source dataframe or ColumnData is given, ydata has to be a list'
                             ' of lists, not of type: {}'.format(type(ydata[0])))
    else:
        xdatad = xdata
        ydatad = ydata

    # draw line plot
    # dataframe and column data source expect all entries to be same length...
    # therefore we parse data to plot routines directly... might make other things harder

    plot_kw_line = plot_params.plot_kwargs(plot_type='line')
    plot_kw_scatter = plot_params.plot_kwargs(plot_type='scatter')
    plot_kw_area = plot_params.plot_kwargs(plot_type='area')

    area_curve = kwargs.pop('area_curve', None)

    for indx, data in enumerate(zip(ydatad, plot_kw_line, plot_kw_scatter, plot_kw_area)):

        yname, kw_line, kw_scatter, kw_area = data

        if isinstance(xdatad, list):
            xdat = xdatad[indx]
        else:
            xdat = xdatad

        if isinstance(source, list):
            sourcet = source[indx]
        else:
            sourcet = source

        if isinstance(default_legend_label, list):
            leg_label = default_legend_label[indx]
        else:
            leg_label = default_legend_label

        if 'legend_label' not in kw_line:
            kw_line['legend_label'] = leg_label
            kw_scatter['legend_label'] = leg_label
            kw_area['legend_label'] = leg_label

        if area_curve is not None:
            if isinstance(area_curve, list):
                try:
                    shift = area_curve[indx]
                except IndexError:
                    shift = area_curve[0]
            else:
                shift = area_curve
        else:
            shift = 0

        if plot_params[('area_plot', indx)]:
            if plot_params[('area_vertical', indx)]:
                p.harea(y=yname, x1=xdat, x2=shift, **kw_area, source=sourcet)
            else:
                p.varea(x=xdat, y1=yname, y2=shift, **kw_area, source=sourcet)

        res = p.line(x=xdat, y=yname, source=sourcet, **kw_line, **kwargs)
        res2 = None
        if plot_points:
            res2 = p.scatter(x=xdat, y=yname, source=sourcet, **kw_scatter)

        if plot_params[('level', indx)] is not None:
            res.level = plot_params[('level', indx)]
            if res2 is not None:
                res2.level = plot_params[('level', indx)]

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.set_legend(p)
    plot_params.save_plot(p)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_dos(dosdata,
              *,
              energy='energy_grid',
              ynames=None,
              energy_label=r'E-E_F [eV]',
              dos_label=r'DOS [1/eV]',
              title=r'Density of states',
              xyswitch=False,
              e_fermi=0,
              outfilename='dos_plot.html',
              **kwargs):
    """
    Create an interactive dos plot (non-spinpolarized) with bokeh
    Both horizontal or vertical orientation are possible

    :param dosdata: source for the dosdata of the plot (pandas Dataframe for example)
    :param energy: key from which to pull the data for the energy grid
    :param ynames: keys from which to pull the data for dos components
    :param energy_label: label for the energy-axis
    :param dos_label: label for the dos-axis
    :param title: title of the figure
    :param xyswitch: bool if True, the energy will be plotted along the y-direction
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param outfilename: filename of the output file

    Kwargs will be passed on to :py:func:`bokeh_line()`
    """

    if 'limits' in kwargs:
        limits = kwargs.pop('limits')
        if xyswitch:
            limits['x'], limits['y'] = limits.pop('dos', None), limits.pop('energy', None)
        else:
            limits['x'], limits['y'] = limits.pop('energy', None), limits.pop('dos', None)
        kwargs['limits'] = {k: v for k, v in limits.items() if v is not None}

    lines = {'horizontal': 0}
    lines['vertical'] = e_fermi

    if xyswitch:
        lines['vertical'], lines['horizontal'] = lines['horizontal'], lines['vertical']

    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             figure_kwargs={
                                 'tooltips': [('Name', '$name'), ('Energy', '@energy_grid{0.0[00]}'),
                                              ('DOS value', '@$name{0.00}')],
                                 'width':
                                 1000,
                             })

    if ynames is None:
        ynames = set(dosdata.keys()) - set([energy] if isinstance(energy, str) else energy)
        ynames = sorted(ynames)

    if xyswitch:
        x, y = ynames, energy
        xlabel, ylabel = dos_label, energy_label
        plot_params.set_defaults(default_type='function', area_vertical=True)
    else:
        xlabel, ylabel = energy_label, dos_label
        x, y = energy, ynames

    p = bokeh_line(dosdata, xdata=x, ydata=y, xlabel=xlabel, ylabel=ylabel, title=title, name=ynames, **kwargs)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_spinpol_dos(dosdata,
                      *,
                      spin_dn_negative=True,
                      energy='energy_grid',
                      ynames=None,
                      energy_label=r'E-E_F [eV]',
                      dos_label=r'DOS [1/eV]',
                      title=r'Density of states',
                      xyswitch=False,
                      e_fermi=0,
                      spin_arrows=True,
                      outfilename='dos_plot.html',
                      **kwargs):
    """
    Create an interactive dos plot (spinpolarized) with bokeh
    Both horizontal or vertical orientation are possible

    :param dosdata: source for the dosdata of the plot (pandas Dataframe for example)
    :param energy: key from which to pull the data for the energy grid
    :param ynames: keys from which to pull the data for dos components
    :param spin_dn_negative: bool, if True (default), the spin down components are plotted downwards
    :param energy_label: label for the energy-axis
    :param dos_label: label for the dos-axis
    :param title: title of the figure
    :param xyswitch: bool if True, the energy will be plotted along the y-direction
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param spin_arrows: bool, if True (default) small arrows will be plotted on the left side of the plot indicating
                        the spin directions (if spin_dn_negative is True)
    :param outfilename: filename of the output file

    Kwargs will be passed on to :py:func:`bokeh_line()`
    """
    from bokeh.models import NumeralTickFormatter, Arrow, NormalHead

    if 'limits' in kwargs:
        limits = kwargs.pop('limits')
        if xyswitch:
            limits['x'], limits['y'] = limits.pop('dos', None), limits.pop('energy', None)
        else:
            limits['x'], limits['y'] = limits.pop('energy', None), limits.pop('dos', None)
        kwargs['limits'] = {k: v for k, v in limits.items() if v is not None}

    lines = {'horizontal': 0}
    lines['vertical'] = e_fermi

    if ynames is None:
        ynames = set(dosdata.keys()) - set([energy] if isinstance(energy, str) else energy)
        ynames = sorted(ynames)
        ynames.extend([f'{key} Spin-Down' for key in ynames])

    if spin_dn_negative:
        dosdata[[key for key in ynames if '_down' in key]] = -dosdata[[key for key in ynames if '_down' in key]]

    if xyswitch:
        lines['vertical'], lines['horizontal'] = lines['horizontal'], lines['vertical']

    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             figure_kwargs={
                                 'tooltips': [('DOS Name', '$name'), ('Energy', '@energy_grid{0.0[00]}'),
                                              ('Value', '@$name{(0,0.00)}')],
                                 'width':
                                 1000
                             })

    if xyswitch:
        x, y = ynames, energy
        xlabel, ylabel = dos_label, energy_label
        plot_params.set_defaults(default_type='function',
                                 area_vertical=True,
                                 x_axis_formatter=NumeralTickFormatter(format='(0,0)'))
    else:
        xlabel, ylabel = energy_label, dos_label
        x, y = energy, ynames
        plot_params.set_defaults(default_type='function',
                                 area_vertical=True,
                                 y_axis_formatter=NumeralTickFormatter(format='(0,0)'))

    plot_params.single_plot = False
    plot_params.num_plots = len(ynames) // 2  #We want the same colors for opposite spin-directions
    plot_params.set_parameters(color=kwargs.pop('color', None), color_palette=kwargs.pop('color_palette', None))
    plot_params.set_color_palette_by_num_plots()

    #Double the colors for spin up and down
    kwargs['color'] = list(plot_params['color'].copy())
    kwargs['color'].extend(kwargs['color'])

    if 'legend_label' not in kwargs:
        kwargs['legend_label'] = list(ynames.copy())
    else:
        if isinstance(kwargs['legend_label'], list):
            if len(kwargs['legend_label']) == len(ynames) // 2:
                kwargs['legend_label'].extend(kwargs['legend_label'])

    if 'show' in kwargs:
        plot_params.set_parameters(show=kwargs['show'])

    with NestedPlotParameters(plot_params):
        p = bokeh_line(dosdata,
                       xdata=x,
                       ydata=y,
                       xlabel=xlabel,
                       ylabel=ylabel,
                       title=title,
                       name=ynames,
                       show=False,
                       **kwargs)

    if spin_arrows and spin_dn_negative:

        #These are hardcoded because the parameters are not
        #reused anywhere (for now)
        x_pos = 50
        length = 70
        pad = 30
        height = p.plot_height - 100
        alpha = 0.5

        p.add_layout(
            Arrow(x_start=x_pos,
                  x_end=x_pos,
                  y_start=height - pad - length,
                  y_end=height - pad,
                  start_units='screen',
                  end_units='screen',
                  line_width=2,
                  line_alpha=alpha,
                  end=NormalHead(line_width=2, size=10, fill_alpha=alpha, line_alpha=alpha)))
        p.add_layout(
            Arrow(x_start=x_pos,
                  x_end=x_pos,
                  y_start=pad + length,
                  y_end=pad,
                  start_units='screen',
                  end_units='screen',
                  line_width=2,
                  line_alpha=alpha,
                  end=NormalHead(line_width=2, size=10, fill_alpha=alpha, line_alpha=alpha)))

    plot_params.save_plot(p)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_bands(bandsdata,
                *,
                k_label='kpath',
                eigenvalues='eigenvalues_up',
                weight=None,
                xlabel='',
                ylabel=r'E-E_F [eV]',
                title='',
                special_kpoints=None,
                size_min=3.0,
                size_scaling=10.0,
                outfilename='bands_plot.html',
                **kwargs):
    """
    Create an interactive bandstructure plot (non-spinpolarized) with bokeh
    Can make a simple plot or weight the size and color of the points against a given weight

    :param bandsdata: source for the bandsdata of the plot (pandas Dataframe for example)
    :param k_label: key from which to pull the data for the kpoints
    :param eigenvalues: key from which to pull the data for eigenvalues
    :param weight: optional key from the bandsdata. If given the size and color of each point
                   are adjusted to show the weights
    :param xlabel: label for the x-axis (default no label)
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param special_kpoints: list of tuples (str, float), place vertical lines at the given values
                            and mark them on the x-axis with the given label
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param size_min: minimum value used in scaling points for weight
    :param size_scaling: factor used in scaling points for weight
    :param outfilename: filename of the output file

    Kwargs will be passed on to :py:func:`bokeh_multi_scatter()`
    """
    from bokeh.transform import linear_cmap

    if weight is not None:

        ylimits = (-15, 15)
        if 'limits' in kwargs:
            if 'y' in kwargs['limits']:
                ylimits = kwargs['limits']['y']

        weight_max = bandsdata[weight].loc[(bandsdata[eigenvalues] > ylimits[0]) &
                                           (bandsdata[eigenvalues] < ylimits[1])].max()

        bandsdata['weight_size'] = size_min + size_scaling * bandsdata[weight] / weight_max
        plot_params.set_defaults(default_type='function',
                                 color=linear_cmap(weight, 'Blues256', weight_max, -0.05),
                                 marker_size='weight_size')
    else:
        plot_params.set_defaults(default_type='function', color='black')

    if special_kpoints is None:
        special_kpoints = []

    xticks = []
    xticklabels = {}
    for label, pos in special_kpoints:
        #if label in ('Gamma', 'g'): Latex label missing for bokeh
        #    label = r'$\Gamma$'
        if pos.is_integer():
            xticklabels[int(pos)] = label
        xticklabels[pos] = label
        xticks.append(pos)

    lines = {'horizontal': 0}
    lines['vertical'] = xticks

    limits = {'y': (-15, 15)}
    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             x_ticks=xticks,
                             x_ticklabels_overwrite=xticklabels,
                             figure_kwargs={
                                 'width': 1280,
                                 'height': 720
                             },
                             x_range_padding=0.0,
                             y_range_padding=0.0,
                             limits=limits)

    return bokeh_multi_scatter(bandsdata,
                               xdata=k_label,
                               ydata=eigenvalues,
                               xlabel='',
                               ylabel=ylabel,
                               title=title,
                               **kwargs)


@ensure_plotter_consistency(plot_params)
def bokeh_spinpol_bands(bandsdata,
                        *,
                        k_label='kpath',
                        eigenvalues=None,
                        weight=None,
                        xlabel='',
                        ylabel=r'E-E_F [eV]',
                        title='',
                        special_kpoints=None,
                        size_min=3.0,
                        size_scaling=10.0,
                        outfilename='bands_plot.html',
                        **kwargs):
    """
    Create an interactive bandstructure plot (spinpolarized) with bokeh
    Can make a simple plot or weight the size and color of the points against a given weight

    :param bandsdata: source for the bandsdata of the plot (pandas Dataframe for example)
    :param k_label: key from which to pull the data for the kpoints
    :param eigenvalues: keys from which to pull the data for eigenvalues (default ['eigenvalues_up','eigenvalues_down'])
    :param weight: optional key from the bandsdata. If given the size and color of each point
                   are adjusted to show the weights
    :param xlabel: label for the x-axis (default no label)
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param special_kpoints: list of tuples (str, float), place vertical lines at the given values
                            and mark them on the x-axis with the given label
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param size_min: minimum value used in scaling points for weight
    :param size_scaling: factor used in scaling points for weight
    :param outfilename: filename of the output file

    Kwargs will be passed on to :py:func:`bokeh_multi_scatter()`
    """
    from bokeh.transform import linear_cmap

    if eigenvalues is None:
        eigenvalues = ['eigenvalues_up', 'eigenvalues_down']

    plot_params.single_plot = False
    plot_params.num_plots = 2

    if weight is not None:
        cmaps = ['Blues256', 'Reds256']
        color = []

        ylimits = (-15, 15)
        if 'limits' in kwargs:
            if 'y' in kwargs['limits']:
                ylimits = kwargs['limits']['y']

        weight_max = bandsdata[weight[0]].loc[(bandsdata[eigenvalues[0]] > ylimits[0]) &
                                              (bandsdata[eigenvalues[0]] < ylimits[1])].max()
        weight_max = max(
            weight_max, bandsdata[weight[1]].loc[(bandsdata[eigenvalues[1]] > ylimits[0]) &
                                                 (bandsdata[eigenvalues[1]] < ylimits[1])].max())

        for indx, (w, cmap) in enumerate(zip(weight, cmaps)):
            color.append(linear_cmap(w, cmap, weight_max, -0.05))
            bandsdata[f'weight_size_{indx}'] = size_min + size_scaling * bandsdata[w] / weight_max
        plot_params.set_defaults(default_type='function', color=color, marker_size=['weight_size_0', 'weight_size_1'])
    else:
        color = ['blue', 'red']
        plot_params.set_defaults(default_type='function', color=color)

    if special_kpoints is None:
        special_kpoints = []

    xticks = []
    xticklabels = {}
    for label, pos in special_kpoints:
        #if label in ('Gamma', 'g'): Latex label missing for bokeh
        #    label = r'$\Gamma$'
        if pos.is_integer():
            xticklabels[int(pos)] = label
        xticklabels[pos] = label
        xticks.append(pos)

    lines = {'horizontal': 0}
    lines['vertical'] = xticks

    limits = {'y': (-15, 15)}
    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             x_ticks=xticks,
                             x_ticklabels_overwrite=xticklabels,
                             figure_kwargs={
                                 'width': 1280,
                                 'height': 720
                             },
                             x_range_padding=0.0,
                             y_range_padding=0.0,
                             limits=limits,
                             level=[None, 'underlay'])

    return bokeh_multi_scatter(bandsdata,
                               xdata=k_label,
                               ydata=eigenvalues,
                               xlabel='',
                               ylabel=ylabel,
                               title=title,
                               **kwargs)


####################################################################################################
##################################### special plots ################################################
####################################################################################################

# periodic table

# tools="pan, xpan, ypan, poly_select, tap, wheel_zoom, xwheel_zoom, ywheel_zoom, xwheel_pan, ywheel_pan,
#         box_zoom, redo, undo, reset, save, crosshair, zoom_out, xzoom_out, yzoom_out, hover"
tooltips_def_period = [('Name', '@name'), ('Atomic number', '@{atomic number}'), ('Atomic mass', '@{atomic mass}'),
                       ('CPK color', '$color[hex, swatch]:CPK'),
                       ('Electronic configuration', '@{electronic configuration}')]


def periodic_table_plot(source,
                        display_values=[],
                        display_positions=[],
                        color_value=None,
                        tooltips=tooltips_def_period,
                        title='',
                        outfilename='periodictable.html',
                        value_color_range=[None, None],
                        log_scale=0,
                        color_map=None,
                        bokeh_palette='Plasma256',
                        toolbar_location=None,
                        tools='hover',
                        blank_color='#c4c4c4',
                        blank_outsiders=[True, True],
                        include_legend=True,
                        copy_source=True,
                        legend_labels=None,
                        color_bar_title=None,
                        show=True):
    """
    Plot function for an interactive periodic table plot. Heat map and hover tool.
    source must be a panda dataframe containing, period, group,

    param source: pandas dataframe containing everything
    param tooltips: what is shown with hover tool. values have to be in source
    example:

    .. code-block:: python

        Keys of panda DF. group, period symbol and atomic number or required...
        Index([u'atomic number', u'symbol', u'name', u'atomic mass', u'CPK',
           u'electronic configuration', u'electronegativity', u'atomic radius',
           u'ion radius', u'van der Waals radius', u'IE-1', u'EA',
           u'standard state', u'bonding type', u'melting point', u'boiling point',
           u'density', u'metal', u'year discovered', u'group', u'period',
           u'rmt_mean', u'rmt_std', u'number_of_occ', u'type_color', u'c_value'],
          dtype='object')

        tooltips_def = [("Name", "@name"),
                    ("Atomic number", "@{atomic number}"),
                    ("Atomic mass", "@{atomic mass}"),
                    ("CPK color", "$color[hex, swatch]:CPK"),
                    ("Electronic configuration", "@{electronic configuration}")]

    param display_values: list of strings, have to match source. Values to be displayed on the element rectangles
    example:["rmt_mean", "rmt_std", "number_of_occ"]
    param display_positions: list of floats, length has to match display_values, At which y offset the display values should be displayed.
    """
    # TODO: solve the use of two different color bars, we just one to use a bokeh color bar and not matplotlib...
    from bokeh.io import export_png
    from bokeh.io import output_notebook, output_file
    from bokeh.sampledata.periodic_table import elements
    from bokeh.transform import dodge, factor_cmap
    from bokeh.models import Arrow, OpenHead, NormalHead, VeeHead
    from bokeh.models import Range1d, LabelSet, Label
    from bokeh.models import LinearColorMapper, LogColorMapper, ColorBar
    from bokeh.models import BasicTicker

    from bokeh.io import show as bshow
    from bokeh.plotting import figure as bokeh_fig
    from matplotlib.colors import Normalize, LogNorm, to_hex
    from matplotlib.cm import plasma  #pylint: disable=no-name-in-module
    from matplotlib.cm import ScalarMappable

    if color_map is None:
        color_map = plasma

    if len(display_values) != len(display_positions):
        raise ValueError(
            'The input lists "display_values" and "display_positions" of "periodic_table_plot" need to have same length.'
        )

    if copy_source:
        source1 = source.copy()
    else:  # inline we change the data here!
        source1 = source

    TOOLTIPS = tooltips

    # defaults
    plot_width = 1470
    plot_height = 1040
    cbar_height = 40
    cbar_width = 500
    cbar_fontsize = 12  # size of cbar labels
    cbar_standoff = 8

    # preprocessing data
    # if colors are not given in source, color
    # if source has "type_color"]

    # colors are tricky, since for the periodic table we use matplotlib, for the legend we have to use bokeh

    # Define color map called 'color_scale'
    data = source1[color_value]
    if value_color_range[0] is not None:
        mind = value_color_range[0]
    else:
        mind = min(data)  # 0.65 #
    if value_color_range[1] is not None:
        maxd = value_color_range[1]
    else:
        maxd = max(data)  # 2.81 #

    if log_scale == 0:
        color_mapper = LinearColorMapper(palette=bokeh_palette, low=mind, high=maxd)
        norm = Normalize(vmin=mind, vmax=maxd)
    elif log_scale == 1:
        for datum in data:
            if datum < 0:
                raise ValueError('Entry for element ' + datum + ' is negative but' ' log-scale is selected')
        color_mapper = LogColorMapper(palette=bokeh_palette, low=mind, high=maxd)
        norm = LogNorm(vmin=mind, vmax=maxd)
    color_scale = ScalarMappable(norm=norm, cmap=color_map).to_rgba(data, alpha=None)

    # Define color for blank entries
    default_value = None
    color_list = []
    color_values = []
    for i in range(len(source1)):
        color_list.append(blank_color)
        color_values.append(default_value)

    for i, data_element in enumerate(source1[color_value]):
        if blank_outsiders[0] and data_element < mind:
            continue
        if blank_outsiders[1] and data_element > maxd:
            continue
        color_list[i] = to_hex(color_scale[i])
        # color_values[i] = data_element

    source1['type_color'] = color_list
    # source["c_value"] = color_values

    if include_legend:
        # we copy the Be entry and display it with some text again at another spot
        be = source1[3:4].copy()
        be['group'] = '7'
        # print(be)
        source1.loc[-1] = be.values[0]
        source1.index = source1.index + 1
        source1 = source1.sort_index()
        # df.head()
    groups = [str(x) for x in range(1, 19)]
    periods = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII']
    # Plot
    p = bokeh_fig(
        title=title,
        plot_width=plot_width,
        plot_height=plot_height,  # 450,
        x_range=groups,
        y_range=list(reversed(periods)),
        tools=tools,
        toolbar_location=toolbar_location,
        tooltips=TOOLTIPS)
    r = p.rect(
        'group',
        'period',
        0.95,
        0.95,
        source=source1,
        fill_alpha=0.6,  # legend="metal",
        color='type_color')

    text_props = {'source': source1, 'text_align': 'left', 'text_baseline': 'middle'}
    x = dodge('group', -0.4, range=p.x_range)

    # The element names
    p.text(x=x,
           y=dodge('period', 0.25, range=p.y_range),
           text='symbol',
           text_font_style='bold',
           text_font_size='14pt',
           **text_props)
    p.text(x=dodge('group', 0.1, range=p.x_range),
           y=dodge('period', 0.25, range=p.y_range),
           text='atomic number',
           text_font_size='12pt',
           **text_props)

    # The values displayed on the element boxes
    for i, label in enumerate(display_values):
        p.text(x=x,
               y=dodge('period', display_positions[i], range=p.y_range),
               text=label,
               text_font_size='10pt',
               **text_props)

        # legend
        # not a real legend, but selfmade text
        # print(be.values[0])
        # print(source.loc(-1))
        # print(be['group'])
        # p.text(x=be['group'], y=dodge("period", 0.25, range=p.y_range), text="test", text_font_style="bold", text_font_size="14pt")
        if legend_labels is not None:
            label1 = legend_labels[i]
        else:
            label1 = label

        # I do not like the hardcoded positions of the legend
        legendlabel = Label(
            x=7.1,
            y=8.4 + display_positions[i],  # x_units='screen', y_units='screen',
            text=label1,
            render_mode='canvas',  # 'css',
            border_line_color='black',
            border_line_alpha=0.0,
            background_fill_color=None,
            background_fill_alpha=1.0)
        p.add_layout(legendlabel)

        legendlabelarrow = Arrow(
            x_start=7.05,
            x_end=6.7,
            y_start=8.5 + display_positions[i],
            y_end=8.5 + display_positions[i],  # x_units='screen', y_units='screen',
            line_width=2,
            end=OpenHead(line_width=2, size=4))  # 'css',
        # border_line_color='black', border_line_alpha=0.0,
        # background_fill_color=None, background_fill_alpha=1.0))
        p.add_layout(legendlabelarrow)

    # labels = LabelSet(x=70, y=80, text=display_values, level='glyph',
    #          x_offset=0, y_offset=5, render_mode='canvas')
    # p.add_layout(labels)

    p.yaxis.major_label_text_font_size = '22pt'
    p.xaxis.visible = False
    p.outline_line_color = None
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_standoff = 0
    # p.legend.orientation = "horizontal"
    p.hover.renderers = [r]
    alpha = 1.0
    # add color bar
    if color_bar_title is None:
        color_bar_title = color_value
    color_bar = ColorBar(
        color_mapper=color_mapper,
        title=color_bar_title,
        title_text_font_size='12pt',
        ticker=BasicTicker(desired_num_ticks=10),
        border_line_color=None,
        background_fill_color=None,
        # 'vertical',
        label_standoff=cbar_standoff,
        location=(plot_width * 0.2, plot_height * 0.69),
        orientation='horizontal',
        scale_alpha=alpha,
        major_label_text_font_size=str(cbar_fontsize) + 'pt',
        height=cbar_height,
        width=cbar_width)

    p.add_layout(color_bar, 'center')

    # deactivate grid

    p.grid.grid_line_color = None

    # export_png(p, filename="plot.png")
    output_file(outfilename)

    if show:
        bshow(p)

    return p


######## a 2d matrix plot ##########

######### plot convergence results plot ########


@ensure_plotter_consistency(plot_params)
def plot_convergence_results(iteration, distance, total_energy, *, show=True, **kwargs):
    """
    Plot the total energy versus the scf iteration
    and plot the distance of the density versus iterations. Uses bokeh_line and bokeh_scatter

    :param iteration: list of Int
    :param distance: list of floats
    :total_energy: list of floats
    :param show: bool, if True call show

    Kwargs will be passed on to :py:func:`bokeh_line()`

    :returns grid: bokeh grid with figures
    """
    from bokeh.layouts import gridplot

    xlabel = r'Iteration'
    ylabel1 = r'Total energy difference [Htr]'
    ylabel2 = r'Distance [me/bohr^3]'
    title1 = r'Total energy difference over scf-Iterations'
    title2 = r'Convergence (log)'

    # since we make a log plot of the total_energy make sure to plot the absolute total energy
    total_energy_abs_diff = []
    for en0, en1 in zip(total_energy[:-1], total_energy[1:]):
        total_energy_abs_diff.append(abs(en1 - en0))

    source1 = pd.DataFrame({'total_energy_delta': total_energy_abs_diff, 'iterations': iteration[1:]})
    source2 = pd.DataFrame({'distance': distance, 'iterations': iteration})

    plot_params.set_defaults(default_type='function',
                             figure_kwargs={
                                 'tools': 'hover,tap,box_zoom,zoom_out,crosshair,reset,save',
                                 'tooltips': [('Iteration', '@x'), ('Total energy distance', '@y')],
                                 'active_inspect': 'hover',
                                 'plot_width': 800,
                                 'plot_height': 450,
                                 'y_axis_type': 'log',
                                 'x_axis_type': 'linear',
                             })

    plot_params.set_parameters(show=show)

    with NestedPlotParameters(plot_params):
        p1 = bokeh_line(source=source1,
                        ydata='total_energy_delta',
                        xdata='iterations',
                        xlabel=xlabel,
                        ylabel=ylabel1,
                        title=title1,
                        name='delta total energy',
                        plot_points=True,
                        show=False,
                        **kwargs)

    plot_params.set_defaults(default_type='function',
                             figure_kwargs={'tooltips': [('Iteration', '@x'), ('Charge distance', '@y')]})

    with NestedPlotParameters(plot_params):
        p2 = bokeh_line(source=source2,
                        ydata='distance',
                        xdata='iterations',
                        xlabel=xlabel,
                        ylabel=ylabel2,
                        title=title2,
                        name='distance',
                        plot_points=True,
                        show=False,
                        **kwargs)

    grid = gridplot([p1, p2], ncols=2)

    plot_params.save_plot(grid)

    return grid


@ensure_plotter_consistency(plot_params)
def plot_convergence_results_m(iterations,
                               distances,
                               total_energies,
                               *,
                               link=False,
                               nodes=None,
                               modes=None,
                               plot_label=None,
                               saveas1='t_energy_convergence',
                               saveas2='distance_convergence',
                               show=True,
                               **kwargs):
    """
    Plot the total energy versus the scf iteration
    and plot the distance of the density versus iterations in a bokeh grid for several SCF results.

    :param distances: list of lists of floats
    :total_energies: list of lists of floats
    :param iterations: list of lists of Int
    :param link: bool, optional default=False:
    :param nodes: list of node uuids or pks important for links
    :param saveas1: str, optional default='t_energy_convergence', save first figure as
    :param saveas2: str, optional default='distance_convergence', save second figure as
    :param figure_kwargs: dict, optional default={'plot_width': 600, 'plot_height': 450}, gets parsed
                          to bokeh_line
    :param kwargs: further key-word arguments for bokeh_line

    :returns grid: bokeh grid with figures
    """
    from bokeh.models import ColumnDataSource
    from bokeh.layouts import gridplot

    xlabel = r'Iteration'
    ylabel1 = r'Total energy difference [Htr]'
    ylabel2 = r'Distance [me/bohr^3]'
    title1 = r'Total energy difference over scf-Iterations'
    title2 = r'Convergence (log)'

    if nodes is not None:
        if not isinstance(nodes, list):
            nodes = [nodes]

    if plot_label is not None:
        if not isinstance(plot_label, list):
            plot_label = [plot_label]

    plot_labels1 = []
    plot_labels2 = []

    data_sources = []
    data_sources2 = []

    tooltips_scatter1 = [('Calculation id', '@id'), ('Iteration', '@x'), ('Total energy difference', '@y')]
    tooltips_scatter2 = [('Calculation id', '@id'), ('Iteration', '@x'), ('Charge distance', '@y')]

    xdata = ['x'] * len(total_energies)
    ydata = ['y'] * len(total_energies)

    # since we make a log plot of the total_energy make sure to plot the absolute total energy
    for i, (iters, total_energy, distance) in enumerate(zip(iterations, total_energies, distances)):
        if nodes is not None:
            node_id = nodes[i]
        else:
            node_id = i

        total_energy_abs_diff = []
        for en0, en1 in zip(total_energy[:-1], total_energy[1:]):
            total_energy_abs_diff.append(abs(en1 - en0))

        plot_labels1.append(f'{node_id}')
        plot_labels2.append(f'{node_id}')
        data = {'y': total_energy_abs_diff, 'x': iters[1:], 'id': [node_id] * len(total_energy_abs_diff)}
        if nodes is not None:
            data['nodes_pk'] = [str(nodes[i])] * len(total_energy_abs_diff)
        if plot_label is not None:
            data['process_label'] = [plot_label[i]] * len(total_energy_abs_diff)

        datasrc = ColumnDataSource(data)
        data_sources.append(datasrc)
        data = {'y': distance, 'x': iters, 'id': [node_id] * len(distance)}
        if nodes is not None:
            data['nodes_pk'] = [str(nodes[i])] * len(distance)
        if plot_label is not None:
            data['process_label'] = [plot_label[i]] * len(distance)

        datasrc = ColumnDataSource(data)
        data_sources2.append(datasrc)

    if plot_label is not None:
        if nodes is None:
            plot_labels1 = plot_label
            plot_labels2 = plot_label
        tooltips_scatter1.append(('process label', '@process_label'))
        tooltips_scatter2.append(('process label', '@process_label'))

    if nodes is not None:
        tooltips_scatter1.append(('outpara pk', '@nodes_pk'))
        tooltips_scatter2.append(('outpara pk', '@nodes_pk'))

    plot_params.set_defaults(default_type='function',
                             figure_kwargs={
                                 'tools': 'hover,tap,box_zoom,zoom_out,crosshair,reset,save,pan',
                                 'tooltips': tooltips_scatter1,
                                 'active_inspect': 'hover',
                                 'plot_width': 800,
                                 'plot_height': 450,
                                 'y_axis_type': 'log',
                                 'x_axis_type': 'linear',
                             },
                             legend_outside_plot_area=True)
    plot_params.set_parameters(show=show)

    # plot
    with NestedPlotParameters(plot_params):
        p1 = bokeh_line(source=data_sources,
                        ydata=ydata,
                        xdata=xdata,
                        xlabel=xlabel,
                        ylabel=ylabel1,
                        title=title1,
                        name=plot_labels1,
                        legend_label=plot_labels1,
                        plot_points=True,
                        show=False,
                        **kwargs)

    plot_params.set_defaults(default_type='function', figure_kwargs={
        'tooltips': tooltips_scatter2,
    })

    with NestedPlotParameters(plot_params):
        p2 = bokeh_line(source=data_sources2,
                        ydata=ydata,
                        xdata=xdata,
                        xlabel=xlabel,
                        ylabel=ylabel2,
                        title=title2,
                        name=plot_labels2,
                        legend_label=plot_labels2,
                        plot_points=True,
                        show=False,
                        **kwargs)

    grid = gridplot([p1, p2], ncols=2)

    plot_params.save_plot(grid)

    return grid


######### plot_convex_hull plot ########


def plot_convex_hull2d(hull,
                       title='Convex Hull',
                       xlabel='Atomic Procentage',
                       ylabel='Formation energy / atom [eV]',
                       linestyle='-',
                       marker='o',
                       legend=True,
                       legend_option={},
                       saveas='convex_hull',
                       limits=[None, None],
                       scale=[None, None],
                       axis=None,
                       color='k',
                       color_line='k',
                       linewidth=2,
                       markersize=8,
                       marker_hull='o',
                       markersize_hull=8,
                       **kwargs):
    """
    Plot method for a 2d convex hull diagram

    :param hull: scipy.spatial.ConvexHull
    """
    pass



# -*- coding: utf-8 -*-

# -*- coding: utf-8 -*-
"""
Tests of the xml_getters
"""
import pytest
import os

#file_path1 = 'files/fleur/aiida_fleur/inpxml'
file_path2 = 'files/fleur/Max-R5'

inpxmlfilefolder = os.path.dirname(os.path.abspath(__file__))

inpxmlfilefolder_valid = [
    #    os.path.abspath(os.path.join(inpxmlfilefolder, file_path1)),
    os.path.abspath(os.path.join(inpxmlfilefolder, file_path2))
]

broken_inputs = [
    'CoHybridPBE0', 'CoUnfold', 'gw1Interface', 'GaAsWannSOC', 'TiO2eelsXML', 'gw2Interface', 'Fe_film_SS_conv',
    'SiHybrid8kpt_nosym', 'SiHybrid8kpt_sym', 'SiHybridGammaNoInv', 'Fe_bulk_SS_conv', 'Fe_film_SSFT',
    'Max-R5/NiO_ldauXML', 'Max-R5/Bi2Te3XML'
]

TEST_FILM_INPXML_PATH = os.path.join(inpxmlfilefolder, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')
TEST_BULK_INPXML_PATH = os.path.join(inpxmlfilefolder, 'files/fleur/Max-R5/SiLOXML/files/inp.xml')
TEST_SINGLE_KPOINT_PATH = os.path.join(inpxmlfilefolder, 'files/fleur/Max-R5/SmAtomjDOS/files/inp.xml')
TEST_MULTIPLE_KPOINT_SETS_PATH = os.path.join(inpxmlfilefolder, 'files/fleur/test_multiple_ksets.xml')
TEST_MAX4_INPXML_PATH = os.path.join(inpxmlfilefolder, 'files/fleur/aiida_fleur/inpxml/FePt/inp.xml')

inpxmlfilelist = []
inpxmlfilelist_content = []
for folder in inpxmlfilefolder_valid:
    for subdir, dirs, files in os.walk(folder):
        for file in files:
            if file.endswith('.xml') and 'inp' in file:
                non_valid = False
                for broken in broken_inputs:
                    if broken in subdir:
                        non_valid = True
                if not non_valid:
                    inpxmlfilelist.append(os.path.join(subdir, file))


@pytest.mark.parametrize('inpxmlfilepath', inpxmlfilelist)
def test_get_cell(load_inpxml, inpxmlfilepath):
    """
    Test that get_cell works for all input files
    """
    from masci_tools.util.xml.xml_getters import get_cell
    import numpy as np

    xmltree, schema_dict = load_inpxml(inpxmlfilepath)

    cell, pbc = get_cell(xmltree, schema_dict)

    assert isinstance(cell, np.ndarray)
    assert cell.shape == (3, 3)
    assert isinstance(pbc, list)
    assert len(pbc) == 3


@pytest.mark.parametrize('inpxmlfilepath', inpxmlfilelist)
def test_get_structure_data(load_inpxml, inpxmlfilepath):
    """
    Test that get_cell works for all input files
    """
    from masci_tools.util.xml.xml_getters import get_structure_data
    import numpy as np

    xmltree, schema_dict = load_inpxml(inpxmlfilepath)

    atoms, cell, pbc = get_structure_data(xmltree, schema_dict)

    assert isinstance(atoms, list)
    assert len(atoms) != 0
    assert isinstance(cell, np.ndarray)
    assert cell.shape == (3, 3)
    assert isinstance(pbc, list)
    assert len(pbc) == 3


@pytest.mark.parametrize('inpxmlfilepath', inpxmlfilelist)
def test_get_parameter_data(load_inpxml, inpxmlfilepath):
    """
    Test that get_cell works for all input files
    """
    from masci_tools.util.xml.xml_getters import get_parameter_data

    xmltree, schema_dict = load_inpxml(inpxmlfilepath)

    para = get_parameter_data(xmltree, schema_dict)

    assert isinstance(para, dict)
    assert para != {}


@pytest.mark.parametrize('inpxmlfilepath', inpxmlfilelist)
def test_get_fleur_modes(load_inpxml, inpxmlfilepath):
    """
    Test that get_cell works for all input files
    """
    from masci_tools.util.xml.xml_getters import get_fleur_modes

    xmltree, schema_dict = load_inpxml(inpxmlfilepath)

    modes = get_fleur_modes(xmltree, schema_dict)

    assert isinstance(modes, dict)
    assert modes != {}


@pytest.mark.parametrize('inpxmlfilepath', inpxmlfilelist)
def test_get_kpoints_data(load_inpxml, inpxmlfilepath):
    """
    Test that get_cell works for all input files
    """
    from masci_tools.util.xml.xml_getters import get_kpoints_data
    import numpy as np

    xmltree, schema_dict = load_inpxml(inpxmlfilepath)

    kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict)

    assert kpoints is not None
    assert weights is not None
    assert isinstance(cell, np.ndarray)
    assert cell.shape == (3, 3)
    assert isinstance(pbc, list)
    assert len(pbc) == 3


def test_get_cell_film(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_cell
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_FILM_INPXML_PATH)

    cell, pbc = get_cell(xmltree, schema_dict)

    data_regression.check({'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_get_cell_bulk(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_cell
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_BULK_INPXML_PATH)

    cell, pbc = get_cell(xmltree, schema_dict)

    data_regression.check({'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_get_structure_film(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_structure_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_FILM_INPXML_PATH)

    atoms, cell, pbc = get_structure_data(xmltree, schema_dict)

    data_regression.check({'atoms': convert_to_pystd(atoms), 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_get_structure_bulk(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_structure_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_BULK_INPXML_PATH)

    atoms, cell, pbc = get_structure_data(xmltree, schema_dict)

    data_regression.check({'atoms': convert_to_pystd(atoms), 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_fleur_modes_film(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_fleur_modes

    xmltree, schema_dict = load_inpxml(TEST_FILM_INPXML_PATH)

    modes = get_fleur_modes(xmltree, schema_dict)

    data_regression.check(modes)


def test_fleur_modes_bulk(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_fleur_modes

    xmltree, schema_dict = load_inpxml(TEST_BULK_INPXML_PATH)

    modes = get_fleur_modes(xmltree, schema_dict)

    data_regression.check(modes)


def test_parameter_film(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_parameter_data

    xmltree, schema_dict = load_inpxml(TEST_FILM_INPXML_PATH)

    para = get_parameter_data(xmltree, schema_dict)

    data_regression.check(para)


def test_parameter_bulk(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_parameter_data

    xmltree, schema_dict = load_inpxml(TEST_BULK_INPXML_PATH)

    para = get_parameter_data(xmltree, schema_dict)

    data_regression.check(para)


def test_kpoints_film(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_kpoints_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_FILM_INPXML_PATH)

    kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict)

    data_regression.check({'kpoints': kpoints, 'weights': weights, 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_kpoints_bulk(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_kpoints_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_BULK_INPXML_PATH)

    kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict)

    data_regression.check({'kpoints': kpoints, 'weights': weights, 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_kpoints_single_kpoint(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_kpoints_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_SINGLE_KPOINT_PATH)

    kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict)

    data_regression.check({'kpoints': kpoints, 'weights': weights, 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_kpoints_multiple_sets(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_kpoints_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_MULTIPLE_KPOINT_SETS_PATH)

    kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict)

    data_regression.check({'kpoints': kpoints, 'weights': weights, 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_kpoints_multiple_sets_selection(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_kpoints_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_MULTIPLE_KPOINT_SETS_PATH)

    kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict, name='default')

    data_regression.check({'kpoints': kpoints, 'weights': weights, 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_kpoints_max4(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_kpoints_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_MAX4_INPXML_PATH)

    kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict)

    data_regression.check({'kpoints': kpoints, 'weights': weights, 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_parameter_max4(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_parameter_data

    xmltree, schema_dict = load_inpxml(TEST_MAX4_INPXML_PATH)

    para = get_parameter_data(xmltree, schema_dict)

    data_regression.check(para)


def test_fleur_modes_max4(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_fleur_modes

    xmltree, schema_dict = load_inpxml(TEST_MAX4_INPXML_PATH)

    modes = get_fleur_modes(xmltree, schema_dict)

    data_regression.check(modes)


def test_get_structure_max4(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_structure_data
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_MAX4_INPXML_PATH)

    atoms, cell, pbc = get_structure_data(xmltree, schema_dict)

    print(type(atoms[0][0][0]))
    data_regression.check({'atoms': convert_to_pystd(atoms), 'cell': convert_to_pystd(cell), 'pbc': pbc})


def test_get_cell_max4(load_inpxml, data_regression):

    from masci_tools.util.xml.xml_getters import get_cell
    from masci_tools.io.common_functions import convert_to_pystd

    xmltree, schema_dict = load_inpxml(TEST_MAX4_INPXML_PATH)

    cell, pbc = get_cell(xmltree, schema_dict)

    data_regression.check({'cell': convert_to_pystd(cell), 'pbc': pbc})



# -*- coding: utf-8 -*-
"""
@author: ruess
"""

# pylint: disable=E0602,W0602

import pytest
from masci_tools.io.parsers.kkrparser_functions import parse_kkr_outputfile, check_error_category


class Test_kkr_parser_functions(object):
    """
    Tests for the kkr parser functions
    """

    grouping_ref = [
        'energy_contour_group', 'warnings_group', 'ewald_sum_group', 'timings_group', 'core_states_group',
        'convergence_group', 'magnetism_group', 'kmesh_group', 'symmetries_group', 'code_info_group'
    ]
    path0 = './files/kkr/kkr_run_slab_soc_simple/'
    outfile = path0 + 'out_kkr'
    outfile_0init = path0 + 'output.0.txt'
    outfile_000 = path0 + 'output.000.txt'
    timing_file = path0 + 'out_timing.000.txt'
    potfile_out = path0 + 'out_potential'
    nonco_out_file = path0 + 'nonco_angle_out.dat'

    def test_complete_kkr_output(self, data_regression):
        """
        Parse complete output of kkr calculation
        """
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, self.outfile, self.outfile_0init, self.outfile_000,
                                                           self.timing_file, self.potfile_out, self.nonco_out_file)
        out_dict['parser_warnings'] = msg_list
        assert success
        assert msg_list == []
        groups = [i for i in list(out_dict.keys()) if 'group' in i]
        assert set(groups) == set(self.grouping_ref)
        data_regression.check(out_dict)

    def test_complete_kkr_output_filehandle(self, data_regression):
        """
        Parse complete output of kkr calculation but using file handles as done in aiida-kkr
        """
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, open(self.outfile), open(self.outfile_0init),
                                                           open(self.outfile_000), open(self.timing_file),
                                                           open(self.potfile_out), open(self.nonco_out_file))
        out_dict['parser_warnings'] = msg_list
        assert success
        assert msg_list == []
        groups = [i for i in list(out_dict.keys()) if 'group' in i]
        assert set(groups) == set(self.grouping_ref)
        data_regression.check(out_dict)

    def test_mag_orbmom_kkr_output(self, data_regression):
        """
        Parse complete output of kkr calculation with orbital moments
        """
        path0 = './files/kkr/kkr_run_slab_soc_mag/'
        outfile = path0 + 'out_kkr'
        outfile_0init = path0 + 'output.0.txt'
        outfile_000 = path0 + 'output.000.txt'
        timing_file = path0 + 'out_timing.000.txt'
        potfile_out = path0 + 'out_potential'
        nonco_out_file = path0 + 'nonco_angle_out.dat'
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, outfile, outfile_0init, outfile_000, timing_file,
                                                           potfile_out, nonco_out_file)
        out_dict['parser_warnings'] = msg_list
        assert success
        assert msg_list == []
        data_regression.check(out_dict)

    def test_nosoc_kkr_output(self, data_regression):
        """
        Parse complete output of kkr calculation nosoc, magnetic
        """

        path0 = './files/kkr/kkr_run_slab_nosoc/'
        outfile = path0 + 'out_kkr'
        outfile_0init = path0 + 'output.0.txt'
        outfile_000 = path0 + 'output.000.txt'
        timing_file = path0 + 'out_timing.000.txt'
        potfile_out = path0 + 'out_potential'
        nonco_out_file = path0 + 'nonco_angle_out.dat'
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, outfile, outfile_0init, outfile_000, timing_file,
                                                           potfile_out, nonco_out_file)
        out_dict['parser_warnings'] = msg_list
        assert success
        assert msg_list == []
        data_regression.check(out_dict)

    def test_missing_outfile(self):
        """
        Parse kkr output where out_kkr is missing. Compares error messages
        """
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, 'wrong_name', self.outfile_0init, self.outfile_000,
                                                           self.timing_file, self.potfile_out, self.nonco_out_file)
        out_dict['parser_warnings'] = msg_list
        assert not success
        assert set(msg_list) == set([
            'Error parsing output of KKR: Version Info', 'Error parsing output of KKR: rms-error',
            'Error parsing output of KKR: charge neutrality', 'Error parsing output of KKR: total magnetic moment',
            'Error parsing output of KKR: spin moment per atom', 'Error parsing output of KKR: orbital moment',
            'Error parsing output of KKR: EF', 'Error parsing output of KKR: DOS@EF',
            'Error parsing output of KKR: total energy', 'Error parsing output of KKR: search for warnings',
            'Error parsing output of KKR: charges', 'Error parsing output of KKR: scfinfo'
        ])

    def test_missing_outfile0init(self):
        """
        Parse kkr output where output.0.txt is missing. Compares error messages
        """
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, self.outfile, 'wrong_name', self.outfile_000,
                                                           self.timing_file, self.potfile_out, self.nonco_out_file)
        out_dict['parser_warnings'] = msg_list
        assert not success
        assert set(msg_list) == set([
            'Error parsing output of KKR: nspin/natom',
            'Error parsing output of KKR: spin moment per atom',
            'Error parsing output of KKR: orbital moment',
            'Error parsing output of KKR: energy contour',
            'Error parsing output of KKR: alat, 2*pi/alat',
            'Error parsing output of KKR: scfinfo',
            'Error parsing output of KKR: kmesh',
            'Error parsing output of KKR: symmetries',
            'Error parsing output of KKR: ewald summation for madelung poterntial',
            'Error parsing output of KKR: lattice vectors (direct/reciprocal)',
            'Error parsing output of KKR: noco angles rms value',
        ])

    def test_missing_outfile000(self):
        """
        Parse kkr output where output.000.txt is missing. Compares error messages
        """
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, self.outfile, self.outfile_0init, 'wrong_name',
                                                           self.timing_file, self.potfile_out, self.nonco_out_file)
        out_dict['parser_warnings'] = msg_list
        assert not success
        assert set(msg_list) == set([
            'Error parsing output of KKR: rms-error', 'Error parsing output of KKR: single particle energies',
            'Error parsing output of KKR: charges', 'Error parsing output of KKR: scfinfo',
            'Error parsing output of KKR: kmesh'
        ])

    def test_missing_timingfile(self):
        """
        Parse kkr output where out_timing.000.txt is missing. Compares error messages
        """
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, self.outfile, self.outfile_0init, self.outfile_000,
                                                           'wrong_name', self.potfile_out, self.nonco_out_file)
        out_dict['parser_warnings'] = msg_list
        assert not success
        assert msg_list == ['Error parsing output of KKR: timings']

    def test_missing_potfile(self):
        """
        Parse kkr output where out_potential is missing. Compares error messages
        """
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, self.outfile, self.outfile_0init, self.outfile_000,
                                                           self.timing_file, 'wrong_name', self.nonco_out_file)
        out_dict['parser_warnings'] = msg_list
        assert not success
        assert msg_list == ['Error parsing output of KKR: core_states']

    def test_missing_nonco_angles(self):
        """
        Parse kkr output where out_potential is missing. Compares error messages
        """
        path0 = './files/kkr/kkr_run_slab_soc_mag/'
        outfile = path0 + 'out_kkr'
        outfile_0init = path0 + 'output.0.txt'
        outfile_000 = path0 + 'output.000.txt'
        timing_file = path0 + 'out_timing.000.txt'
        potfile_out = path0 + 'out_potential'
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, outfile, outfile_0init, outfile_000, timing_file,
                                                           potfile_out, 'wrong_name')
        out_dict['parser_warnings'] = msg_list
        assert not success
        assert msg_list == ['Error parsing output of KKR: spin moment per atom']

    def test_check_error_category(self):
        """
        Check check_error_category function used in parser after parse_kkr_outputfile is used
        """
        fname = 'nonco_angles_out.dat'
        err_cat, err_msg = (2, 'Error! NONCO_ANGLES_OUT not found {}'.format(fname))
        assert not check_error_category(err_cat, err_msg, {'use_newsosol': False})
        assert check_error_category(err_cat, err_msg, {'use_newsosol': True})

    def test_parse_dosout(self, data_regression):
        """
        Parse output of dos calculation since ouput changes slightly (e.g. no ewald sum)
        """
        path0 = './files/kkr/kkr_run_dos_output/'
        outfile = path0 + 'out_kkr'
        outfile_0init = path0 + 'output.0.txt'
        outfile_000 = path0 + 'output.000.txt'
        timing_file = path0 + 'out_timing.000.txt'
        potfile_out = path0 + 'out_potential'
        out_dict = {}
        success, msg_list, out_dict = parse_kkr_outputfile(out_dict, outfile, outfile_0init, outfile_000, timing_file,
                                                           potfile_out, 'wrong_name')
        out_dict['parser_warnings'] = msg_list
        assert success
        assert msg_list == []
        data_regression.check(out_dict)

    def test_parse_3Dsymmetries(self, data_regression):
        """
        Parse output of a dos calculation in 3D (used to fail due to symmetries reading)
        """
        p = './files/kkr/parser_3Dsymmetries/'
        success, msg_list, out_dict = parse_kkr_outputfile({}, p + 'out_kkr', p + 'output.0.txt', p + 'output.000.txt',
                                                           p + 'out_timing.000.txt', p + 'out_potential',
                                                           p + 'nonco_angle_out.dat')
        assert success
        assert msg_list == []
        data_regression.check(out_dict)

    def test_Nan_output(self):
        """
        Parse output of a dos calculation in 3D (used to fail due to symmetries reading)
        """
        p = './files/kkr/parser_3Dsymmetries/'
        success, msg_list, out_dict = parse_kkr_outputfile({}, p + 'out_kkr', p + 'output.0.txt', p + 'output.000.txt',
                                                           p + 'out_timing.000.txt', p + 'out_potential',
                                                           p + 'nonco_angle_out.dat', p + 'output.2.txt')
        from numpy import isnan
        captured_nan = False
        for key, val in out_dict['convergence_group'].items():
            if key in ['charge_neutrality', 'rms']:
                if isnan(val):
                    captured_nan = True
            elif key in [
                    'charge_neutrality_all_iterations', 'dos_at_fermi_energy_all_iterations',
                    'fermi_energy_all_iterations', 'rms_all_iterations', 'total_energy_Ry_all_iterations',
                    'rms_per_atom'
            ]:
                for isub in val:
                    if isnan(isub):
                        captured_nan = True
        assert success
        assert not captured_nan



# -*- coding: utf-8 -*-
"""
Tests of the crystal field calculations
"""
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import pytest


def test_CFCalculation_txt_files():
    """
    Test of the CFCalculation reading the data from txt files
    """
    from masci_tools.tools.cf_calculation import CFCalculation, CFCoefficient

    #Make sure new script produces the same result as old one
    expected_results = [
        CFCoefficient(l=2,
                      m=0,
                      spin_up=-419.7891726292168,
                      spin_down=-414.7152560307904,
                      unit='K',
                      convention='Stevens'),
        CFCoefficient(l=4,
                      m=0,
                      spin_up=-35.92607948104669,
                      spin_down=-26.384951772020756,
                      unit='K',
                      convention='Stevens'),
        CFCoefficient(l=6, m=0, spin_up=6.522900740505054, spin_down=5.488104692050172, unit='K', convention='Stevens')
    ]

    cf = CFCalculation(reference_radius='cdn')
    cf.readPot('files/cf_calculation/VKS.2.0.dat',
               'files/cf_calculation/VKS.4.0.dat',
               'files/cf_calculation/VKS.6.0.dat',
               lm=[(2, 0), (4, 0), (6, 0)])
    cf.readCDN('files/cf_calculation/Nd.dat', header=3)
    cf.cdn['RMT'] = 3.138049652
    results = cf.performIntegration()

    assert results == expected_results


def test_CFCalculation_hdf_files():
    """
    Test of the CFCalculation using the hdf file produced by fleur
    """
    from masci_tools.tools.cf_calculation import CFCalculation, CFCoefficient

    expected_results = [
        CFCoefficient(l=2, m=0, spin_up=-571.68845386399, spin_down=-558.2336974657351, unit='K', convention='Stevens'),
        CFCoefficient(l=4,
                      m=0,
                      spin_up=-34.982539807305045,
                      spin_down=-21.850435868549834,
                      unit='K',
                      convention='Stevens'),
        CFCoefficient(l=6, m=0, spin_up=3.8503494779930776, spin_down=2.168215129491561, unit='K',
                      convention='Stevens'),
        CFCoefficient(l=6,
                      m=-6,
                      spin_up=110.50156137060345,
                      spin_down=85.58558990378205,
                      unit='K',
                      convention='Stevens'),
        CFCoefficient(l=6, m=6, spin_up=110.50156137060345, spin_down=85.58558990378205, unit='K', convention='Stevens')
    ]

    cf = CFCalculation()
    cf.readPot('files/cf_calculation/CFdata.hdf')
    cf.readCDN('files/cf_calculation/CFdata.hdf')
    results = cf.performIntegration()

    assert results == expected_results


def test_CFCalculation_hdf_files_wybourne_convention():
    """
    Test of the CFCalculation using the hdf file produced by fleur
    """
    from masci_tools.tools.cf_calculation import CFCalculation, CFCoefficient

    expected_results = [
        CFCoefficient(l=2,
                      m=0,
                      spin_up=(-1143.37690772798 + 0j),
                      spin_down=(-1116.4673949314702 + 0j),
                      unit='K',
                      convention='Wybourne'),
        CFCoefficient(l=4,
                      m=0,
                      spin_up=(-279.86031845844036 + 0j),
                      spin_down=(-174.80348694839867 + 0j),
                      unit='K',
                      convention='Wybourne'),
        CFCoefficient(l=6,
                      m=0,
                      spin_up=(61.60559164788924 + 0j),
                      spin_down=(34.69144207186498 + 0j),
                      unit='K',
                      convention='Wybourne'),
        CFCoefficient(l=6,
                      m=-6,
                      spin_up=(116.32750335918315 + 4.696327749935313e-06j),
                      spin_down=(90.09789430612014 + 3.6373963939901583e-06j),
                      unit='K',
                      convention='Wybourne'),
        CFCoefficient(l=6,
                      m=6,
                      spin_up=(116.32750335918315 - 4.696327749935313e-06j),
                      spin_down=(90.09789430612014 - 3.6373963939901583e-06j),
                      unit='K',
                      convention='Wybourne')
    ]

    cf = CFCalculation()
    cf.readPot('files/cf_calculation/CFdata.hdf')
    cf.readCDN('files/cf_calculation/CFdata.hdf')
    results = cf.performIntegration(convert=False)

    print(results)
    assert results == expected_results


@pytest.mark.mpl_image_compare(baseline_dir='files/cf_calculation/', filename='cf_calculation.png')
def test_plot_crystal_field_calculation():
    """
    Test of the plot illustrating the potential and charge density going into the calculation
    """
    from masci_tools.tools.cf_calculation import CFCalculation, plot_crystal_field_calculation

    cf = CFCalculation()
    cf.readPot('files/cf_calculation/CFdata.hdf')
    cf.readCDN('files/cf_calculation/CFdata.hdf')

    plt.gcf().clear()

    plot_crystal_field_calculation(cf, show=False)

    return plt.gcf()


@pytest.mark.mpl_image_compare(baseline_dir='files/cf_calculation/', filename='cf_potential.png')
def test_plot_crystal_field_potential():
    """
    Test of the plot illustraing the resulting crystal field potential
    """
    from masci_tools.tools.cf_calculation import CFCoefficient, plot_crystal_field_potential

    coeffs = [
        CFCoefficient(l=2,
                      m=0,
                      spin_up=(-1143.37690772798 + 0j),
                      spin_down=(-1116.4673949314702 + 0j),
                      unit='K',
                      convention='Wybourne'),
        CFCoefficient(l=4,
                      m=0,
                      spin_up=(-279.86031845844036 + 0j),
                      spin_down=(-174.80348694839867 + 0j),
                      unit='K',
                      convention='Wybourne'),
        CFCoefficient(l=6,
                      m=0,
                      spin_up=(61.60559164788924 + 0j),
                      spin_down=(34.69144207186498 + 0j),
                      unit='K',
                      convention='Wybourne'),
        CFCoefficient(l=6,
                      m=-6,
                      spin_up=(116.32750335918315 + 4.696327749935313e-06j),
                      spin_down=(90.09789430612014 + 3.6373963939901583e-06j),
                      unit='K',
                      convention='Wybourne'),
        CFCoefficient(l=6,
                      m=6,
                      spin_up=(116.32750335918315 - 4.696327749935313e-06j),
                      spin_down=(90.09789430612014 - 3.6373963939901583e-06j),
                      unit='K',
                      convention='Wybourne')
    ]

    plt.gcf().clear()
    plot_crystal_field_potential(coeffs, show=False)

    return plt.gcf()



# -*- coding: utf-8 -*-
"""
Configurations for masci_tools tests
"""
import pytest


@pytest.fixture
def load_inpxml():
    """Returns the etree and schema_dict generator"""

    def _load_inpxml(path):
        from lxml import etree
        from masci_tools.io.parsers.fleur.fleur_schema import InputSchemaDict
        with open(path, 'r') as inpxmlfile:
            tree = etree.parse(inpxmlfile)
            version = str(tree.xpath('//@fleurInputVersion')[0])
            schema_dict = InputSchemaDict.fromVersion(version)
        return tree, schema_dict

    return _load_inpxml


@pytest.fixture
def clean_parser_log():
    """Clean parser logs for consistent testing"""

    def _clean_parser_log(info_dict):

        clean_dict = {}
        clean_dict['parser_info'] = info_dict.pop('parser_info', [])
        clean_dict['parser_warnings'] = info_dict.pop('parser_warnings', [])
        clean_dict['parser_errors'] = info_dict.pop('parser_errors', [])

        return clean_dict

    return _clean_parser_log


@pytest.fixture(scope='session', autouse=True)
def disable_parser_tracebacks():
    """Disable logging of tracebacks in parser logs Thanks to
       https://stackoverflow.com/questions/54605699/python-logging-disable-stack-trace"""
    import logging

    class TracebackInfoFilter(logging.Filter):
        """Clear or restore the exception on log records"""

        def filter(self, record):
            record._exc_info_hidden, record.exc_info = record.exc_info, None
            # clear the exception traceback text cache, if created.
            record.exc_text = None
            return True

    inp_logger = logging.getLogger('masci_tools.io.parsers.fleur.fleur_inpxml_parser')
    out_logger = logging.getLogger('masci_tools.io.parsers.fleur.fleur_outxml_parser')

    traceback_filter = TracebackInfoFilter()
    inp_logger.addFilter(traceback_filter)
    out_logger.addFilter(traceback_filter)

    yield  #Now all tests run

    inp_logger.removeFilter(traceback_filter)
    out_logger.removeFilter(traceback_filter)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module provides functions to extract distinct parts of the fleur xml files
for easy versioning and reuse
"""
from masci_tools.io.parsers.fleur.fleur_schema import schema_dict_version_dispatch


def get_fleur_modes(xmltree, schema_dict):

    from masci_tools.util.schema_dict_util import read_constants
    from masci_tools.util.schema_dict_util import evaluate_attribute, tag_exists
    from masci_tools.util.xml.common_functions import clear_xml

    xmltree = clear_xml(xmltree)
    root = xmltree.getroot()
    constants = read_constants(root, schema_dict)

    fleur_modes = {}
    fleur_modes['jspin'] = evaluate_attribute(root, schema_dict, 'jspins', constants=constants)

    noco = evaluate_attribute(root, schema_dict, 'l_noco', constants=constants, optional=True)
    if noco is None:
        noco = False
    fleur_modes['noco'] = noco

    soc = evaluate_attribute(root, schema_dict, 'l_soc', constants=constants, optional=True)
    if soc is None:
        soc = False
    fleur_modes['soc'] = soc

    forces = evaluate_attribute(root, schema_dict, 'l_f', constants=constants, optional=True)
    if forces is None:
        forces = False
    fleur_modes['forces'] = forces

    gw = evaluate_attribute(root, schema_dict, 'gw', constants=constants, optional=True)
    if gw is None:
        gw = False
    else:
        gw = gw != 0
    fleur_modes['gw'] = gw

    if schema_dict.inp_version > (0, 27):
        fleur_modes['force_theorem'] = tag_exists(root, schema_dict, 'forceTheorem')
    else:
        fleur_modes['force_theorem'] = False

    fleur_modes['film'] = tag_exists(root, schema_dict, 'filmPos')
    fleur_modes['ldau'] = tag_exists(root, schema_dict, 'ldaU', contains='species')
    fleur_modes['dos'] = evaluate_attribute(root, schema_dict, 'dos', constants=constants)
    fleur_modes['band'] = evaluate_attribute(root, schema_dict, 'band', constants=constants)
    fleur_modes['bz_integration'] = evaluate_attribute(root,
                                                       schema_dict,
                                                       'mode',
                                                       constants=constants,
                                                       tag_name='bzIntegration')

    return fleur_modes


def get_cell(xmltree, schema_dict):

    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_text, tag_exists
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.util.xml.converters import convert_xml_attribute
    from masci_tools.util.constants import BOHR_A
    import numpy as np

    xmltree = clear_xml(xmltree)

    root = xmltree.getroot()

    constants = read_constants(root, schema_dict)

    cell = None
    lattice_tag = None
    if tag_exists(root, schema_dict, 'bulkLattice'):
        lattice_tag = eval_simple_xpath(root, schema_dict, 'bulkLattice')
        pbc = [True, True, True]
    elif tag_exists(root, schema_dict, 'filmLattice'):
        lattice_tag = eval_simple_xpath(root, schema_dict, 'filmLattice')
        pbc = [True, True, False]

    if lattice_tag is not None:
        row1 = evaluate_text(lattice_tag,
                             schema_dict,
                             'row-1',
                             constants=constants,
                             contains='bravaisMatrix',
                             optional=True)
        row2 = evaluate_text(lattice_tag,
                             schema_dict,
                             'row-2',
                             constants=constants,
                             contains='bravaisMatrix',
                             optional=True)
        row3 = evaluate_text(lattice_tag,
                             schema_dict,
                             'row-3',
                             constants=constants,
                             contains='bravaisMatrix',
                             optional=True)

        if all(x is not None and x != [] for x in [row1, row2, row3]):
            #Explicit Conversion to float for versions Max4 and before
            if schema_dict.inp_version < (0, 33):
                row1, suc = convert_xml_attribute(row1, ['float_expression'], constants=constants)
                row2, suc = convert_xml_attribute(row2, ['float_expression'], constants=constants)
                row3, suc = convert_xml_attribute(row3, ['float_expression'], constants=constants)

            cell = np.array([row1, row2, row3]) * BOHR_A

    if cell is None:
        raise ValueError('Could not extract Bravais matrix out of inp.xml. Is the '
                         'Bravais matrix explicitly given? i.e Latnam definition '
                         'not supported.')

    return cell, pbc


def get_parameter_data(xmltree, schema_dict, inpgen_ready=True, write_ids=True):
    """
    This routine returns an python dictionary produced from the inp.xml
    file, which can be used as a calc_parameters node by inpgen.
    Be aware that inpgen does not take all information that is contained in an inp.xml file

    :param inpxmlfile: and xml etree of a inp.xml file
    :param inpgen_ready: Bool, return a dict which can be inputed into inpgen while setting atoms
    :return new_parameters: A Dict, which will lead to the same inp.xml (in case if other defaults,
                            which can not be controlled by input for inpgen, were changed)

    """
    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_attribute, evaluate_text
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.util.xml.converters import convert_fleur_lo, convert_xml_attribute
    from masci_tools.io.common_functions import filter_out_empty_dict_entries

    # TODO: convert econfig
    # TODO: parse kpoints, somehow count is bad (if symmetry changes), mesh is not known, path cannot be specified

    ########
    parameters = {}

    xmltree = clear_xml(xmltree)
    root = xmltree.getroot()
    constants = read_constants(root, schema_dict)

    # Create the cards

    # &input # most things are not needed for AiiDA here. or we ignor them for now.
    # film is set by the plugin depended on the structure
    # symor per default = False? to avoid input which fleur can't take

    # &comp
    comp_dict = {}
    comp_dict['jspins'] = evaluate_attribute(root, schema_dict, 'jspins', constants=constants)
    comp_dict['frcor'] = evaluate_attribute(root, schema_dict, 'frcor', constants=constants, optional=True)
    comp_dict['ctail'] = evaluate_attribute(root, schema_dict, 'ctail', constants=constants)
    comp_dict['kcrel'] = evaluate_attribute(root, schema_dict, 'kcrel', constants=constants, optional=True)
    comp_dict['gmax'] = evaluate_attribute(root, schema_dict, 'gmax', constants=constants)
    comp_dict['gmaxxc'] = evaluate_attribute(root, schema_dict, 'gmaxxc', constants=constants)
    comp_dict['kmax'] = evaluate_attribute(root, schema_dict, 'kmax', constants=constants)

    if schema_dict.inp_version <= (0, 31):
        comp_dict['gmax'], _ = convert_xml_attribute(comp_dict['gmax'], ['float', 'float_expression'],
                                                     constants=constants)
        comp_dict['gmaxxc'], _ = convert_xml_attribute(comp_dict['gmaxxc'], ['float', 'float_expression'],
                                                       constants=constants)
        comp_dict['kmax'], _ = convert_xml_attribute(comp_dict['kmax'], ['float', 'float_expression'],
                                                     constants=constants)

    parameters['comp'] = filter_out_empty_dict_entries(comp_dict)

    # &atoms
    species_list = eval_simple_xpath(root, schema_dict, 'species', list_return=True)
    species_several = {}
    # first we see if there are several species with the same atomic number
    for species in species_list:
        atom_z = evaluate_attribute(species, schema_dict, 'atomicNumber', constants)
        species_several[atom_z] = species_several.get(atom_z, 0) + 1

    species_count = {}
    for indx, species in enumerate(species_list):
        atom_dict = {}
        atoms_name = 'atom{}'.format(indx)
        atom_z = evaluate_attribute(species, schema_dict, 'atomicNumber', constants)
        if not inpgen_ready:
            atom_dict['z'] = atom_z
        species_count[atom_z] = species_count.get(atom_z, 0) + 1
        atom_id = f'{atom_z}.{species_count[atom_z]}'
        if write_ids:
            if species_several[atom_z] > 1:
                atom_dict['id'] = atom_id

        if schema_dict.inp_version <= (0, 31):
            atom_dict['ncst'] = evaluate_attribute(species, schema_dict, 'coreStates', constants)
        atom_dict['rmt'] = evaluate_attribute(species, schema_dict, 'radius', constants=constants)
        atom_dict['dx'] = evaluate_attribute(species, schema_dict, 'logIncrement', constants=constants)
        atom_dict['jri'] = evaluate_attribute(species, schema_dict, 'gridPoints', constants=constants)
        atom_dict['lmax'] = evaluate_attribute(species, schema_dict, 'lmax', constants=constants)
        atom_dict['lnonsph'] = evaluate_attribute(species, schema_dict, 'lnonsphr', constants=constants)
        atom_dict['bmu'] = evaluate_attribute(species, schema_dict, 'magMom', constants, optional=True)

        atom_dict['element'] = evaluate_attribute(species, schema_dict, 'element', constants=constants)

        #atom_econfig = eval_simple_xpath(species, schema_dict, 'electronConfig')
        atom_lo = eval_simple_xpath(species, schema_dict, 'lo', list_return=True)
        #atom_econfig = eval_simple_xpath(species, schema_dict, 'electronConfig')

        if len(atom_lo) != 0:
            atom_dict['lo'] = convert_fleur_lo(atom_lo)

        if schema_dict.inp_version <= (0, 31):
            atom_dict['bmu'], _ = convert_xml_attribute(atom_dict['bmu'], ['float', 'float_expression'],
                                                        constants=constants)
            atom_dict['dx'], _ = convert_xml_attribute(atom_dict['dx'], ['float', 'float_expression'],
                                                       constants=constants)
            atom_dict['rmt'], _ = convert_xml_attribute(atom_dict['rmt'], ['float', 'float_expression'],
                                                        constants=constants)

        parameters[atoms_name] = filter_out_empty_dict_entries(atom_dict)

    # &soc
    soc = evaluate_attribute(root, schema_dict, 'l_soc', constants=constants, optional=True)
    theta = evaluate_attribute(root, schema_dict, 'theta', constants=constants, contains='soc', optional=True)
    phi = evaluate_attribute(root, schema_dict, 'phi', constants=constants, contains='soc', optional=True)
    if soc is not None and soc:
        if schema_dict.inp_version <= (0, 31):
            theta, _ = convert_xml_attribute(theta, ['float', 'float_expression'], constants=constants)
            phi, _ = convert_xml_attribute(phi, ['float', 'float_expression'], constants=constants)

        parameters['soc'] = {'theta': theta, 'phi': phi}

    # &kpt
    #attrib = convert_from_fortran_bool(eval_xpath(root, l_soc_xpath))
    #theta = eval_xpath(root, theta_xpath)
    #phi = eval_xpath(root, phi_xpath)
    # if kpt:
    #    new_parameters['kpt'] = {'theta' : theta, 'phi' : phi}
    #    # ['nkpt', 'kpts', 'div1', 'div2', 'div3',                         'tkb', 'tria'],

    # title
    title = evaluate_text(root, schema_dict, 'comment', constants=constants, optional=True)
    if title:
        parameters['title'] = title.replace('\n', '').strip()

    # &exco
    #TODO, easy
    exco_dict = {}
    exco_dict['xctyp'] = evaluate_attribute(root, schema_dict, 'name', constants, contains='xcFunctional')
    # 'exco' : ['xctyp', 'relxc'],
    parameters['exco'] = filter_out_empty_dict_entries(exco_dict)
    # &film
    # TODO

    # &qss
    # TODO

    # lattice, not supported?

    return parameters


def get_structure_data(xmltree, schema_dict):

    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_text, evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.util.xml.converters import convert_xml_attribute
    from masci_tools.io.common_functions import rel_to_abs, rel_to_abs_f

    xmltree = clear_xml(xmltree)

    root = xmltree.getroot()

    constants = read_constants(root, schema_dict)

    cell, pbc = get_cell(xmltree, schema_dict)

    species_names = evaluate_attribute(root, schema_dict, 'name', constants=constants, contains='species')
    species_elements = evaluate_attribute(root, schema_dict, 'element', constants=constants, contains='species')

    if not isinstance(species_names, list):
        species_names = [species_names]
    if not isinstance(species_elements, list):
        species_elements = [species_elements]

    if len(species_names) != len(species_elements):
        raise ValueError(
            f'Failed to read in species names and elements. Got {len(species_names)} names and {len(species_elements)} elements'
        )

    species_dict = dict(zip(species_names, species_elements))

    atom_data = []
    atom_groups = eval_simple_xpath(root, schema_dict, 'atomGroup', list_return=True)
    for group in atom_groups:

        group_species = evaluate_attribute(group, schema_dict, 'species', constants=constants)

        atom_positions = []

        absolute_positions = evaluate_text(group,
                                           schema_dict,
                                           'absPos',
                                           constants=constants,
                                           list_return=True,
                                           optional=True)
        relative_positions = evaluate_text(group,
                                           schema_dict,
                                           'relPos',
                                           constants=constants,
                                           list_return=True,
                                           optional=True)
        film_positions = evaluate_text(group,
                                       schema_dict,
                                       'filmPos',
                                       constants=constants,
                                       list_return=True,
                                       optional=True)

        if schema_dict.inp_version < (0, 33):
            for indx, pos in enumerate(absolute_positions):
                absolute_positions[indx], suc = convert_xml_attribute(pos, ['float', 'float_expression'],
                                                                      constants=constants)
            for indx, pos in enumerate(relative_positions):
                relative_positions[indx], suc = convert_xml_attribute(pos, ['float', 'float_expression'],
                                                                      constants=constants)
            for indx, pos in enumerate(film_positions):
                film_positions[indx], suc = convert_xml_attribute(pos, ['float', 'float_expression'],
                                                                  constants=constants)

        atom_positions = absolute_positions

        for rel_pos in relative_positions:
            atom_positions.append(rel_to_abs(rel_pos, cell))

        for film_pos in film_positions:
            atom_positions.append(rel_to_abs_f(film_pos, cell))

        if len(atom_positions) == 0:
            raise ValueError('Failed to read atom positions for group')

        atom_data.extend((pos, species_dict[group_species]) for pos in atom_positions)

    return atom_data, cell, pbc


@schema_dict_version_dispatch(output_schema=False)
def get_kpoints_data(xmltree, schema_dict, name=None):

    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_text, evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.util.xml.converters import convert_xml_attribute

    xmltree = clear_xml(xmltree)

    root = xmltree.getroot()

    constants = read_constants(root, schema_dict)

    cell, pbc = get_cell(xmltree, schema_dict)

    kpointlists = eval_simple_xpath(root, schema_dict, 'kPointList', list_return=True)

    if len(kpointlists) == 0:
        raise ValueError('No Kpoint lists found in the given inp.xml')

    labels = [kpoint_set.attrib.get('name') for kpoint_set in kpointlists]
    if name is not None and name not in labels:
        raise ValueError(f'Found no Kpoint list with the name: {name}' f'Available list names: {labels}')

    kpoints_data = {}
    weights_data = {}
    for kpointlist in kpointlists:

        label = evaluate_attribute(kpointlist, schema_dict, 'name')

        if name is not None and name != label:
            continue

        kpoints = evaluate_text(kpointlist, schema_dict, 'kPoint', constants=constants, list_return=True)
        weights = evaluate_attribute(kpointlist, schema_dict, 'weight', constants=constants, list_return=True)

        if schema_dict.inp_version == (0, 32):
            for indx, kpoint in enumerate(kpoints):
                kpoints[indx], suc = convert_xml_attribute(kpoint, ['float', 'float_expression'], constants=constants)
            weights, suc = convert_xml_attribute(weights, ['float', 'float_expression'],
                                                 constants=constants,
                                                 list_return=True)

        if not isinstance(kpoints[0], list):
            kpoints = [kpoints]
            weights = [weights]

        kpoints_data[label] = kpoints
        weights_data[label] = weights

    if len(kpoints_data) == 1:
        _, kpoints_data = kpoints_data.popitem()
        _, weights_data = weights_data.popitem()

    return kpoints_data, weights_data, cell, pbc


@get_kpoints_data.register(max_version='0.31')
def get_kpoints_data_max4(xmltree, schema_dict):

    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_text, evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.util.xml.converters import convert_xml_attribute

    xmltree = clear_xml(xmltree)

    root = xmltree.getroot()

    constants = read_constants(root, schema_dict)

    cell, pbc = get_cell(xmltree, schema_dict)

    kpointlist = eval_simple_xpath(root, schema_dict, 'kPointList', list_return=True, not_contains='altKPoint')

    if len(kpointlist) == 0:
        raise ValueError('No Kpoint lists found in the given inp.xml')

    kpointlist = kpointlist[0]

    kpoints = evaluate_text(kpointlist,
                            schema_dict,
                            'kPoint',
                            constants=constants,
                            not_contains='altKPoint',
                            list_return=True)
    weights = evaluate_attribute(kpointlist,
                                 schema_dict,
                                 'weight',
                                 constants=constants,
                                 not_contains='altKPoint',
                                 list_return=True)

    for indx, kpoint in enumerate(kpoints):
        kpoints[indx], suc = convert_xml_attribute(kpoint, ['float', 'float_expression'], constants=constants)
    weights, suc = convert_xml_attribute(weights, ['float', 'float_expression'], constants=constants, list_return=True)

    return kpoints, weights, cell, pbc



# -*- coding: utf-8 -*-
"""
Test for the FleurXMLModifier calss
"""
import os
import pytest

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
TEST_INPXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')
TEST_INPXML_LDAU_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/GaAsMultiUForceXML/files/inp.xml')
TEST_NMMPMAT_PATH = os.path.join(FILE_PATH, 'files/fleur/input_nmmpmat.txt')


def test_fleurxmlmodifier_facade_methods():
    """
    Make sure that adding all facade methods results in the right task list
    """
    from masci_tools.io.fleurxmlmodifier import FleurXMLModifier, ModifierTask

    fm = FleurXMLModifier()

    actions = fm.get_avail_actions()

    assert fm._tasks == []

    for action in actions.values():
        action('TEST_ARG', random_kwarg='TEST2')

    assert len(actions) == len(fm._tasks)

    for task, action in zip(fm._tasks, actions.values()):
        assert isinstance(task, ModifierTask)
        assert task.name == action.__name__
        assert task.args == ('TEST_ARG',)
        assert task.kwargs == {'random_kwarg': 'TEST2'}


def test_fleurxml_modifier_modify_xmlfile_simple():
    """Tests if fleurinp_modifier with various modifations on species"""
    from masci_tools.io.fleurxmlmodifier import FleurXMLModifier, ModifierTask

    fm = FleurXMLModifier()
    fm.set_inpchanges({'dos': True, 'Kmax': 3.9})
    fm.shift_value({'Kmax': 0.1}, 'rel')
    fm.shift_value_species_label('                 222', 'radius', 3, mode='abs')
    fm.set_species('all', {'mtSphere': {'radius': 3.333}})
    fm.xml_set_attrib_value_no_create('/fleurInput/calculationSetup/cutoffs', 'Gmax', '14.0')

    assert fm.changes() == [
        ModifierTask(name='set_inpchanges', args=({
            'dos': True,
            'Kmax': 3.9
        },), kwargs={}),
        ModifierTask(name='shift_value', args=({
            'Kmax': 0.1
        }, 'rel'), kwargs={}),
        ModifierTask(name='shift_value_species_label',
                     args=(
                         '                 222',
                         'radius',
                         3,
                     ),
                     kwargs={'mode': 'abs'}),
        ModifierTask(name='set_species', args=('all', {
            'mtSphere': {
                'radius': 3.333
            }
        }), kwargs={}),
        ModifierTask(name='xml_set_attrib_value_no_create',
                     args=(
                         '/fleurInput/calculationSetup/cutoffs',
                         'Gmax',
                         '14.0',
                     ),
                     kwargs={})
    ]

    #The underlying methods are tested in the specific tests for the setters
    #We only want to ensure that the procedure finishes without error
    xmltree = fm.modify_xmlfile(TEST_INPXML_PATH)

    assert xmltree is not None


def test_fleurxml_modifier_modify_xmlfile_undo():
    """Tests if fleurinp_modifier with various modifations on species"""
    from masci_tools.io.fleurxmlmodifier import FleurXMLModifier, ModifierTask

    fm = FleurXMLModifier()
    fm.set_inpchanges({'dos': True, 'Kmax': 3.9})
    fm.shift_value({'Kmax': 0.1}, 'rel')
    fm.shift_value_species_label('                 222', 'radius', 3, mode='abs')
    fm.set_species('all', {'mtSphere': {'radius': 3.333}})

    fm.undo()

    assert fm.changes() == [
        ModifierTask(name='set_inpchanges', args=({
            'dos': True,
            'Kmax': 3.9
        },), kwargs={}),
        ModifierTask(name='shift_value', args=({
            'Kmax': 0.1
        }, 'rel'), kwargs={}),
        ModifierTask(name='shift_value_species_label',
                     args=(
                         '                 222',
                         'radius',
                         3,
                     ),
                     kwargs={'mode': 'abs'})
    ]

    #The underlying methods are tested in the specific tests for the setters
    #We only want to ensure that the procedure finishes without error
    xmltree = fm.modify_xmlfile(TEST_INPXML_PATH)

    assert xmltree is not None


def test_fleurxml_modifier_modify_xmlfile_undo_revert_all():
    """Tests if fleurinp_modifier with various modifations on species"""
    from masci_tools.io.fleurxmlmodifier import FleurXMLModifier

    fm = FleurXMLModifier()
    fm.set_inpchanges({'dos': True, 'Kmax': 3.9})
    fm.shift_value({'Kmax': 0.1}, 'rel')
    fm.shift_value_species_label('                 222', 'radius', 3, mode='abs')
    fm.set_species('all', {'mtSphere': {'radius': 3.333}})

    fm.undo(revert_all=True)

    assert fm.changes() == []

    #The underlying methods are tested in the specific tests for the setters
    #We only want to ensure that the procedure finishes without error
    xmltree = fm.modify_xmlfile(TEST_INPXML_PATH)

    assert xmltree is not None


def test_fleurxmlmodifier_nmmpmat():
    """Tests if set_nmmpmat works on fleurinp modifier works, with right interface"""
    from masci_tools.io.fleurxmlmodifier import FleurXMLModifier

    fm = FleurXMLModifier()
    fm.set_nmmpmat('Ga-1', orbital=2, spin=1, state_occupations=[1, 2, 3, 4, 5])
    fm.set_nmmpmat('As-2', orbital=1, spin=1, denmat=[[1, -2, 3], [4, -5, 6], [7, -8, 9]])

    # Does not validate
    # Found invalid diagonal element for species Ga-1, spin 1 and l=2
    with pytest.raises(ValueError, match=r'Changes were not valid \(n_mmp_mat file is not compatible\)'):
        fm.modify_xmlfile(TEST_INPXML_LDAU_PATH)
    xmltree, nmmpmat = fm.modify_xmlfile(TEST_INPXML_LDAU_PATH, validate_changes=False)

    assert xmltree is not None
    assert nmmpmat is not None

    xmltree, nmmpmat = fm.modify_xmlfile(TEST_INPXML_LDAU_PATH,
                                         original_nmmp_file=TEST_NMMPMAT_PATH,
                                         validate_changes=False)

    assert xmltree is not None
    assert nmmpmat is not None



# -*- coding: utf-8 -*-
"""
Tests of the out.xml parser for Fleur
"""
import pytest
from masci_tools.io.parsers.fleur import outxml_parser
import os
import math
from pprint import pprint

# Collect the input files
file_path = 'files/fleur/Max-R5'

outxmlfilefolder = os.path.dirname(os.path.abspath(__file__))
outxmlfilefolder_valid = [os.path.abspath(os.path.join(outxmlfilefolder, file_path))]

outxmlfilelist = []
for folder in outxmlfilefolder_valid:
    for subdir, dirs, files in os.walk(folder):
        for file in files:
            if file.endswith('.xml') and 'out' in file:
                outxmlfilelist.append(os.path.join(subdir, file))


@pytest.mark.parametrize('outxmlfilepath', outxmlfilelist)
def test_outxml_valid_outxml(outxmlfilepath):
    """
    test if valid inp.xml files are recognized by the inpxml_parser
    """
    from lxml import etree

    #Pass outxmlfile
    out_dict = outxml_parser(outxmlfilepath)

    assert out_dict is not None
    assert isinstance(out_dict, dict)
    assert out_dict != {}

    #Parse before
    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(outxmlfilepath, parser)
    out_dict = outxml_parser(xmltree)

    assert out_dict is not None
    assert isinstance(out_dict, dict)
    assert out_dict != {}

    #call with contextmanager
    with open(outxmlfilepath, 'r') as outfile:
        out_dict = outxml_parser(outfile)

    assert out_dict is not None
    assert isinstance(out_dict, dict)
    assert out_dict != {}


def test_outxml_validation_errors(data_regression, clean_parser_log):
    """
    Test the output parser against files for detecting validation
    """

    OUTXML_FILEPATH1 = os.path.join(outxmlfilefolder, 'files/fleur/broken_out_xml/simple_validation_error.xml')

    with pytest.raises(ValueError, match='Output file does not validate against the schema:'):
        out_dict = outxml_parser(OUTXML_FILEPATH1)

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH1, ignore_validation=True, parser_info_out=warnings)

    data_regression.check({'output_dict': out_dict, 'warnings': clean_parser_log(warnings)})


def test_outxml_empty_out(data_regression, clean_parser_log):
    """
    Test the output parser against empty file
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder, 'files/fleur/broken_out_xml/empty_out.xml')

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings)

    data_regression.check({'output_dict': out_dict, 'warnings': clean_parser_log(warnings)})


def test_outxml_broken(data_regression, clean_parser_log):
    """
    Test the output parser against a file which terminates after some iterations
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder, 'files/fleur/broken_out_xml/terminated.xml')

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings)

    data_regression.check({
        'output_dict': out_dict,
        'warnings': clean_parser_log(warnings),
    })


def test_outxml_broken_firstiter(data_regression, clean_parser_log):
    """
    Test the output parser against a file which terminates in the first iteration
    """
    OUTXML_FILEPATH = os.path.join(outxmlfilefolder, 'files/fleur/broken_out_xml/terminated_firstit.xml')
    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings)

    data_regression.check({
        'output_dict': out_dict,
        'warnings': clean_parser_log(warnings),
    })


def test_outxml_garbage_values(data_regression, clean_parser_log):
    """
    Test the behaviour of the ouput parser when encountering NaN, Inf or fortran formatting errors ****
    """
    OUTXML_FILEPATH = os.path.join(outxmlfilefolder, 'files/fleur/broken_out_xml/garbage_values.xml')

    def isNaN(num):
        return math.isnan(num)  #num != num

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, ignore_validation=True, parser_info_out=warnings)

    assert isNaN(out_dict['fermi_energy'])
    assert out_dict['magnetic_moments'] == '********'
    assert out_dict['total_charge'] == float('Inf')

    data_regression.check({'warnings': clean_parser_log(warnings)})


def test_outxml_incompatible_versions():
    """
    Test the output parser against files with broken/wrong or unsupported version strings
    """

    #output version does not exist
    OUTXML_FILEPATH1 = os.path.join(outxmlfilefolder, 'files/fleur/broken_out_xml/non_existing_version.xml')
    with pytest.raises(FileNotFoundError, match='No FleurOutputSchema.xsd found'):
        out_dict = outxml_parser(OUTXML_FILEPATH1)

    #version string 0.27 and programVersion='fleur 27' not supported
    OUTXML_FILEPATH1 = os.path.join(outxmlfilefolder, 'files/fleur/broken_out_xml/non_supported_version.xml')
    with pytest.raises(ValueError, match="Unknown fleur version: File-version '0.27' Program-version 'fleur 20'"):
        out_dict = outxml_parser(OUTXML_FILEPATH1)


def test_outxml_invalid_iteration():
    """
    Test the output parser with invaid iteration to parse arguments
    """

    #output version does not exist (InputSchema is loaded first so this is the raised error)
    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')
    with pytest.raises(ValueError, match=r"Valid values are: 'first', 'last', 'all', or int"):
        out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse=('Test', 3))
        pprint(out_dict)

    with pytest.raises(ValueError, match=r"Got '999'; but only '6' iterations are available"):
        out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse=999)


def test_outxml_additional_tasks_simple(data_regression):
    """
    Test the definition of additional tasks (resding an attribute)
    """
    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    TEST_TASK_ITERATION = {
        'core_states': {
            'core_eig_val_sum': {
                'parse_type': 'attrib',
                'path_spec': {
                    'name': 'eigValSum'
                }
            }
        }
    }

    out_dict = outxml_parser(OUTXML_FILEPATH, additional_tasks=TEST_TASK_ITERATION)

    data_regression.check({
        'output_dict': out_dict,
    })

    TEST_TASK_ITERATION_INVALID = {
        'core_states': {
            'core_eig_val_sum': {
                'parse_type': 'attrib',
                'path_spec': {
                    'name': 'eigValSum'
                },
                'ignore': ['eigValSum']
            }
        }
    }
    with pytest.raises(ValueError, match="Got extra Keys: {'ignore'}"):
        out_dict = outxml_parser(OUTXML_FILEPATH, additional_tasks=TEST_TASK_ITERATION_INVALID)


def test_outxml_additional_tasks_allattribs(data_regression):
    """
    Test the definition of additional tasks (reading an all attributes of a tag)
    """
    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    TEST_TASK_ITERATION_ALLATTRIBS = {
        'core_states': {
            'core_info': {
                'parse_type': 'allAttribs',
                'path_spec': {
                    'name': 'coreStates'
                },
                'subdict': 'core_info'
            }
        }
    }
    out_dict = outxml_parser(OUTXML_FILEPATH, additional_tasks=TEST_TASK_ITERATION_ALLATTRIBS)
    data_regression.check({
        'output_dict': out_dict,
    })

    TEST_TASK_ITERATION_ALLATTRIBS_INVALID = {
        'core_states': {
            'core_info': {
                'parse_type': 'allAttribs',
                'path_spec': {
                    'name': 'coreStates'
                },
                'overwrite_last': True
            }
        }
    }
    with pytest.raises(ValueError, match="Got extra Keys: {'overwrite_last'}"):
        out_dict = outxml_parser(OUTXML_FILEPATH, additional_tasks=TEST_TASK_ITERATION_ALLATTRIBS_INVALID)


def test_outxml_add_tasks_overwrite(data_regression):
    """
    Test the overwriting of tasks
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    REPLACE_BANDGAP = {
        'bandgap': {
            'bandgap': {
                'parse_type': 'singleValue',
                'path_spec': {
                    'name': 'freeEnergy'
                },
                'only_required': True
            }
        }
    }

    with pytest.raises(ValueError, match="Task 'bandgap' is already defined."):
        out_dict = outxml_parser(OUTXML_FILEPATH, additional_tasks=REPLACE_BANDGAP)

    out_dict = outxml_parser(OUTXML_FILEPATH, additional_tasks=REPLACE_BANDGAP, overwrite=True)
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_add_tasks_append(data_regression):
    """
    Test the append option for defining additional tasks
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    #Move the number_of_atom_types from general_out_info to general_inp_info
    #and write the comment from the inp.xml into it
    #This tests both the correct insertin in general_inp_info and that inner keys can be
    #overwritten in general_out_info

    REPLACE_DICT = {
        'general_out_info': {
            'number_of_atom_types': {}
        },
        'general_inp_info': {
            'number_of_atom_types': {
                'parse_type': 'text',
                'path_spec': {
                    'name': 'comment'
                }
            }
        }
    }

    with pytest.raises(ValueError, match="Task 'general_out_info' is already defined."):
        out_dict = outxml_parser(OUTXML_FILEPATH, additional_tasks=REPLACE_DICT)

    out_dict = outxml_parser(OUTXML_FILEPATH, additional_tasks=REPLACE_DICT, append=True)
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_pre_max3_1compatibility(data_regression, clean_parser_log):
    """
    Test if older than Max3.1 output files are processed correctly (and a warning should be shown for this case)
    """

    OUTXML_FILEPATH = os.path.abspath(os.path.join(outxmlfilefolder, 'files/fleur/old_versions/Max3_0_test_out.xml'))

    warnings = {}
    with pytest.warns(UserWarning):
        out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings, iteration_to_parse='all')
    data_regression.check({
        'output_dict': out_dict,
        'warnings': clean_parser_log(warnings),
    })


def test_outxml_max3_1compatibility(data_regression, clean_parser_log):
    """
    Test if Max3.1 output files are processed correctly
    """

    OUTXML_FILEPATH = os.path.abspath(os.path.join(outxmlfilefolder, 'files/fleur/old_versions/Max3_1_test_out.xml'))

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings, iteration_to_parse='all')
    data_regression.check({
        'output_dict': out_dict,
        'warnings': clean_parser_log(warnings),
    })


def test_outxml_max4compatibility(data_regression, clean_parser_log):
    """
    Test if Max4 output files are processed correctly
    """

    OUTXML_FILEPATH = os.path.abspath(os.path.join(outxmlfilefolder, 'files/fleur/old_versions/Max4_test_out.xml'))

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings, iteration_to_parse='all')
    data_regression.check({
        'output_dict': out_dict,
        'warnings': clean_parser_log(warnings),
    })


def test_outxml_max5_0_compatibility(data_regression, clean_parser_log):
    """
    Test if Max5.0 output files are processed correctly
    """

    OUTXML_FILEPATH = os.path.abspath(os.path.join(outxmlfilefolder, 'files/fleur/old_versions/Max5_0_test_out.xml'))

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings, iteration_to_parse='all')
    data_regression.check({
        'output_dict': out_dict,
        'warnings': clean_parser_log(warnings),
    })


def test_outxml_differing_versions(data_regression, clean_parser_log):
    """
    Test if files with different input/output versions are parsed correctly
    """
    OUTXML_FILEPATH = os.path.abspath(os.path.join(outxmlfilefolder, 'files/fleur/output_mixed_versions.xml'))

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings)

    data_regression.check({
        'output_dict': out_dict,
        'warnings': clean_parser_log(warnings),
    })


def test_outxml_newer_version(data_regression, clean_parser_log):
    """
    Test if files with not yet existent versions are parsed correctly (fallback to last available)
    """
    OUTXML_FILEPATH = os.path.abspath(os.path.join(outxmlfilefolder, 'files/fleur/output_newer_version.xml'))

    warnings = {}
    out_dict = outxml_parser(OUTXML_FILEPATH, parser_info_out=warnings)

    data_regression.check({
        'output_dict': out_dict,
        'warnings': clean_parser_log(warnings),
    })


def test_outxml_lastiter(data_regression):
    """
    Test the parsing of only the last iteration
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    out_dict = outxml_parser(OUTXML_FILEPATH)
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_firstiter(data_regression):
    """
    Test the parsing of only the first iteration
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse='first')
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_alliter(data_regression):
    """
    Test the parsing of all available iterations
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse='all')
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_indexiter(data_regression):
    """
    Test the parsing of an iteration specified by index
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse=3)
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_minimal_mode(data_regression):
    """
    Test the minimal mode of the outxml_parser
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'SiLOXML/files/out.xml')

    out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse='all', minimal_mode=True)
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_magnetic(data_regression):
    """
    Test the outxml_parser for magnetic calculations
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'Fe_bct_LOXML/files/out.xml')

    out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse='all')
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_ldaurelax(data_regression):
    """
    Test the outxml_parser for LDA+U and forces
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'GaAsMultiUForceXML/files/out.xml')

    out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse='all')
    data_regression.check({
        'output_dict': out_dict,
    })


def test_outxml_force(data_regression):
    """
    Test the outxml_parser for a forcetheorem calculation
    """

    OUTXML_FILEPATH = os.path.join(outxmlfilefolder_valid[0], 'FePt_film_SSFT_LO/files/out.xml')

    out_dict = outxml_parser(OUTXML_FILEPATH, iteration_to_parse='all')

    data_regression.check({
        'output_dict': out_dict,
    })



# -*- coding: utf-8 -*-
"""
@author: ruess
"""

import pytest
from masci_tools.io.parsers.voroparser_functions import parse_voronoi_output


class Test_voronoi_parser_functions(object):
    """
    Tests for the voronoi parser functions
    """

    grouping_ref = ['volumes_group', 'radii_atoms_group', 'code_info_group', 'core_states_group', 'cluster_info_group']

    path0 = '../tests/files/voronoi/'
    outfile = path0 + 'out_voronoi'
    potfile = path0 + 'output.pot'
    atominfo = path0 + 'atominfo.txt'
    radii = path0 + 'radii.dat'
    inputfile = path0 + 'inputcard'

    def test_complete_voro_output(self, data_regression):
        """
        Parse complete output of voronoi calculation and compare out_dict, grouping, warnings
        """
        out_dict = {'parser_version': 'some_version_number'}
        success, msg_list, out_dict = parse_voronoi_output(out_dict, self.outfile, self.potfile, self.atominfo,
                                                           self.radii, self.inputfile)
        out_dict['parser_warnings'] = msg_list

        assert success
        assert msg_list == []
        groups = [i for i in list(out_dict.keys()) if 'group' in i]
        assert set(groups) == set(self.grouping_ref)

        data_regression.check(out_dict)

    def test_complete_voro_output_filehandle(self, data_regression):
        """
        Parse complete output of voronoi calculation from open file handles as done in aiida-kkr and compare out_dict, grouping, warnings
        """
        out_dict = {'parser_version': 'some_version_number'}
        success, msg_list, out_dict = parse_voronoi_output(out_dict, open(self.outfile), open(self.potfile),
                                                           open(self.atominfo), open(self.radii), open(self.inputfile))
        out_dict['parser_warnings'] = msg_list
        assert success
        assert msg_list == []
        groups = [i for i in list(out_dict.keys()) if 'group' in i]
        assert set(groups) == set(self.grouping_ref)

        data_regression.check(out_dict)

    def test_missing_outfile(self, data_regression):
        """
        Parse output where out_voronoi is missing and compare error messages/rest of out_dict
        """
        out_dict = {'parser_version': 'some_version_number'}
        success, msg_list, out_dict = parse_voronoi_output(out_dict, '', self.potfile, self.atominfo, self.radii,
                                                           self.inputfile)
        out_dict['parser_warnings'] = msg_list

        assert not success
        data_regression.check({'msg_list': msg_list, 'output': out_dict})

    def test_missing_atominfo(self, data_regression):
        """
        Parse output where atominfo.txt is missing and compare error messages/rest of out_dict
        """
        out_dict = {'parser_version': 'some_version_number'}
        success, msg_list, out_dict = parse_voronoi_output(out_dict, self.outfile, self.potfile, 'wrong_name',
                                                           self.radii, self.inputfile)
        out_dict['parser_warnings'] = msg_list
        assert not success
        data_regression.check({'msg_list': msg_list, 'output': out_dict})

    def test_missing_inputfile(self, data_regression):
        """
        Parse output where inputcard is missing and compare error messages/rest of out_dict
        """
        out_dict = {'parser_version': 'some_version_number'}
        success, msg_list, out_dict = parse_voronoi_output(out_dict, self.outfile, self.potfile, self.atominfo,
                                                           self.radii, 'wrong_name')
        out_dict['parser_warnings'] = msg_list
        assert not success
        data_regression.check({'msg_list': msg_list, 'output': out_dict})

    def test_missing_potfile(self, data_regression):
        """
        Parse output where output.pot is missing and compare error messages/rest of out_dict
        """
        out_dict = {'parser_version': 'some_version_number'}
        success, msg_list, out_dict = parse_voronoi_output(out_dict, self.outfile, 'wrong_name', self.atominfo,
                                                           self.radii, self.inputfile)
        out_dict['parser_warnings'] = msg_list

        assert not success
        data_regression.check({'msg_list': msg_list, 'output': out_dict})

    def test_missing_radii(self, data_regression):
        """
        Parse output where radii.dat is missing and compare error messages/rest of out_dict
        """
        out_dict = {'parser_version': 'some_version_number'}
        success, msg_list, out_dict = parse_voronoi_output(out_dict, self.outfile, self.potfile, self.atominfo,
                                                           'wrong_name', self.inputfile)
        out_dict['parser_warnings'] = msg_list

        assert not success
        data_regression.check({'msg_list': msg_list, 'output': out_dict})
